from modules.module import Module, debug_decorator
import numpy as np
from utils.terrains import Terrain


class Combine(Module):
    ''' Combined terrains '''
    create_folder = False

    @debug_decorator
    def __call__(self, operation='Add', terrain_temp=None, terrain_prim=None,
                 default=None, last=None, **_):
        '''
        Combine terrains in terrain_temp (or terrain_prim) using 'operation'

        Args:
          operation (string): What operation to perform
          last (int): Use last N terrains. If None, then all are used

        '''
        # Initialize lists
        terrain_temp = [] if terrain_temp is None else terrain_temp
        terrain_prim = [] if terrain_prim is None else terrain_prim

        operations = ['Add', 'Max', 'Min', 'Prod']
        operation = default if default is not None else operation
        assert operation in operations, f'operation not in {operations}'

        if last is not None and last > 0:
            # When e.g. taking the last two, the slice is [-2:]
            # So here we make sure this is negative
            last = -last

        from utils.utils import get_terrains
        terrains = get_terrains(
            terrain_temp, terrain_prim, last=last, remove=True,
            print_fn=self.info)

        # Turn into arrays. Workaround to work in Windows/Ubuntu
        terrain_arrays = [terrain.array for terrain in terrains]

        # Apply any of the operations
        if operation == 'Add':
            array = np.sum(terrain_arrays, axis=0)
            terrain = Terrain.from_array(
                array, size=terrains[0].size, extent=terrains[0].extent)
        elif operation == 'Max':
            array = np.maximum.reduce(terrain_arrays)
            terrain = Terrain.from_array(
                array, size=terrains[0].size, extent=terrains[0].extent)
        elif operation == 'Min':
            array = np.minimum.reduce(terrain_arrays)
            terrain = Terrain.from_array(
                array, size=terrains[0].size, extent=terrains[0].extent)
        elif operation == 'Prod':
            array = np.prod(terrain_arrays, axis=0)
            terrain = Terrain.from_array(
                array, size=terrains[0].size, extent=terrains[0].extent)
        else:
            raise AttributeError

        # Add terrain to heap
        terrain_prim.append(terrain)

        # Return updated heap/dict
        return {
            'terrain_temp': terrain_temp,
            'terrain_prim': terrain_prim,
        }


class CombineLast(Combine):
    ''' Combine, with only the last two terrains,

    'terrain' and the latest from the 'heap'
    '''
    def __call__(self, *args, last=None, **kwargs):
        return super(CombineLast, self).__call__(*args, last=2, **kwargs)


class ToPrimary(Combine):
    ''' Move terrains to primary '''
    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=[], terrain_prim=[],
                 default=None, last=None, **_):
        from utils.utils import get_terrains
        terrain = get_terrains(terrain_temp, terrain_prim, last,
                               print_fn=self.info)
        terrain_prim.extend(terrain)

        return {
            'terrain_temp': terrain_temp,
            'terrain_prim': terrain_prim,
        }


class WeightedSum(Module):
    ''' Weighted sum of terrains '''
    create_folder = False

    @debug_decorator
    def __call__(self, weights=[5, 8, 0.1], terrain_temp=None,
                 terrain_prim=None, last=None,
                 overwrite=False,
                 default=None, **_):

        # Initialize lists
        terrain_temp = [] if terrain_temp is None else terrain_temp
        terrain_prim = [] if terrain_prim is None else terrain_prim

        weights = default if default is not None else weights
        weights = np.array(weights).reshape(-1, 1, 1)

        self.debug(f"Use weights:{weights.squeeze()}")

        # Get terrains
        from utils.utils import get_terrains
        terrains = get_terrains(
            terrain_temp, terrain_prim, last=last, remove=True,
            print_fn=self.info)

        # Get arrays
        terrain_arrays = [terrain.array for terrain in terrains]

        array = np.sum(np.multiply(terrain_arrays, weights), axis=0)
        terrain = Terrain.from_array(
            array, size=terrains[0].size, extent=terrains[0].extent)

        terrain_prim.append(terrain)

        return {
            'terrain_temp': terrain_temp,
            'terrain_prim': terrain_prim,
            }


class Compose(Module):
    ''' Combined two (and only) terrains using a mask (factor) + operation '''
    create_folder = False

    @debug_decorator
    def __call__(self, operation='Over',
                 terrain_temp=None, terrain_prim=None,
                 factor=None, default=None, last=None, **_):
        '''
        Combine terrains two terrains using some operation and mask (factor)

        Args:
          operation (string): What operation to perform
        '''
        # Initialize lists
        terrain_temp = [] if terrain_temp is None else terrain_temp
        terrain_prim = [] if terrain_prim is None else terrain_prim

        operations = ['Over', 'Under', 'OverAndAlign', 'UnderAndAlign']
        operation = default if default is not None else operation
        assert operation in operations, f'{operation} not in {operations}'

        from utils.utils import get_terrains
        terrains = get_terrains(
            terrain_temp, terrain_prim, last=2, remove=True,
            print_fn=self.info)

        # Turn into arrays. Workaround to work in Windows/Ubuntu
        array_0 = terrains[0].array
        array_1 = terrains[1].array

        # clip factor to [0, 1]
        factor = np.clip(factor, 0, 1)

        # Standard alpha blend: top terrain over base
        if operation == 'Over':
            array = array_1*factor + array_0*(1-factor)
        # Inverted alpha blend: base terrain over top
        if operation == 'Under':
            array = array_0*factor + array_1*(1-factor)
        # Blend over, then shift to align mean height
        if operation == 'OverAndAlign':
            difference_in_mean = np.sum(array_0*factor - array_1*factor)/np.sum(factor)
            array = array_1*factor + array_0*(1-factor) + difference_in_mean*factor
        # Blend under, then shift to align mean height
        if operation == 'UnderAndAlign':
            difference_in_mean = np.sum(array_1*factor - array_0*factor) / np.sum(factor)
            array = array_0*factor + array_1*(1-factor) + difference_in_mean*factor

        terrain = Terrain.from_array(
            array, size=terrains[0].size, extent=terrains[0].extent)

        # Add terrain to heap
        terrain_prim.append(terrain)

        # Return updated primary/temporary
        return {
            'terrain_temp': terrain_temp,
            'terrain_prim': terrain_prim,
        }


class Stack(Module):
    ''' Place terrains side by side '''
    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=None, terrain_prim=None,
                 default=None, last=None, **_):
        '''
        Operates on the 'most recent' list.
        '''
        # Initialize lists
        terrain_temp = [] if terrain_temp is None else terrain_temp
        terrain_prim = [] if terrain_prim is None else terrain_prim

        if last is not None and last > 0:
            # When e.g. taking the last two, the slice is [-2:]
            # So here we make sure this is negative
            last = -last

        from utils.utils import get_terrains
        terrains = get_terrains(
            terrain_temp, terrain_prim, last=last, remove=True,
            print_fn=self.info)

        from utils.terrains import assign_grid_indices
        from utils.terrains import merge_terrain_blocks

        # Construct grid-indices from terrain extents
        grid_indices = assign_grid_indices(terrains)
        index_to_terrain = {}
        for grid_index, terrain in zip(grid_indices, terrains):
            index_to_terrain[grid_index] = [terrain]

        # Add the merged terrain to the primary list
        merged_terrain_list = merge_terrain_blocks(index_to_terrain)
        terrain_prim.extend(merged_terrain_list)

        # We need to return these in case we have emptited some list
        # (made it a new list?), as then the existing list is not 'updated'
        return {
            'terrain_temp': terrain_temp,
            'terrain_prim': terrain_prim,
        }



import numpy as np
import yaml


class Obstacles:
    # array_types = ['position', 'radius', 'height']
    array_types = ['position', 'width', 'height', 'yaw_deg', 'aspect',
                   'pitch_deg']
    limit_name_mapping = {'position': 'pos_lim', 'width': 'width_lim'}

    def __init__(self, Y=None, N=None, **kwargs):
        '''
        Obstacles

        Args:
          Y: terrain classification in [0, 5]
          position_distribution: normalized 2D np.array
        '''
        if Y is not None:
            # Set using Y and position_distribution
            self.set_from_Y(Y=Y, **kwargs)
        elif N is not None:
            # Set using N, pos_lim, rad_lim
            self.set_from_limits(N, **kwargs)

    def __len__(self):
        return self.position.shape[1]

    def __getitem__(self, key):
        cls = type(self)
        # If slice, apply it to the arrays (second axis),
        # else get single index, but make sure too keep dims
        if isinstance(key, (slice, np.ndarray)):
            dict_of_arrays = {
                'position': self.position[:, key],
                'width': self.width[:, key],
                'height': self.height[:, key],
                'yaw_deg': self.yaw_deg[:, key],
                'aspect': self.aspect[:, key],
                'pitch_deg': self.pitch_deg[:, key],
            }
            return cls.from_arrays(dict_of_arrays)
        else:
            dict_of_arrays = {
                'position': self.position[:, key, np.newaxis],
                'width': self.width[:, key, np.newaxis],
                'height': self.height[:, key, np.newaxis],
                'yaw_deg': self.yaw_deg[:, key, np.newaxis],
                'aspect': self.aspect[:, key, np.newaxis],
                'pitch_deg': self.pitch_deg[:, key, np.newaxis],
            }
            return cls.from_arrays(dict_of_arrays)

    def __setitem__(self, position, piece):
        # self.pieces[position] = piece
        raise NotImplementedError

    def __add__(self, other):
        # Construct new combined arrays
        dict_of_new_arrays = {
            'position': np.concatenate((self.position, other.position), axis=1),
            'width': np.concatenate((self.width, other.width), axis=1),
            'height': np.concatenate((self.height, other.height), axis=1),
            'yaw_deg': np.concatenate((self.yaw_deg, other.yaw_deg), axis=1),
            'aspect': np.concatenate((self.aspect, other.aspect), axis=1),
            'pitch_deg': np.concatenate((self.pitch_deg, other.pitch_deg), axis=1),
        }
        cls = type(self)

        return cls.from_arrays(dict_of_new_arrays)

    def generator(self):
        for i in range(len(self)):
            position = self.position[:, i].tolist()
            width = self.width[0, i]
            height = self.height[0, i]
            yaw_deg = self.yaw_deg[0, i]
            aspect = self.aspect[0, i]
            pitch_deg = self.pitch_deg[0, i]

            yield position, width, height, yaw_deg, aspect, pitch_deg

    def __iter__(self):
        return self.generator()

    def __str__(self):
        # Return a string representation of the object
        return self.__repr__()

    def __repr__(self):
        # Return a complete string representation of the object
        repr_str = f"Obstacles(N={len(self)}"

        for name in self.array_types:
            if name in self.limit_name_mapping:
                array = getattr(self, name)
                limit_name = self.limit_name_mapping[name]
                left = [float(f"{a:.5g}") for a in np.min(array, axis=1)]
                right = [float(f"{a:.5g}") for a in np.max(array, axis=1)]
                if len(left) == 1:
                    left = left[0]
                if len(right) == 1:
                    right = right[0]

                repr_str += f", {limit_name}=[{left}, {right}]"

        # Add final parenthesis
        repr_str += ")"

        return repr_str

    def set_from_dict_of_arrays(self, dict_of_arrays):
        for name, array in dict_of_arrays.items():
            setattr(self, name, array)

    def set_from_Y(self, Y=3, size=(50, 50), position_distribution=None):
        '''


        Args:
          size: used for uniform distribution, unless other specified
          position_distribution: optional positional distribution
        '''
        from utils.utils import generate_size_distribution
        from utils.utils import spawn_particles
        # Generate density mapping
        obstacle_density_mapping = generate_size_distribution(Y=Y)
        # Uniform position distribution
        if position_distribution is None:
            position_distribution = np.ones(size)/np.multiply(*size)

        # Spawn particles
        dict_of_arrays = spawn_particles(
            obstacle_density_mapping, position_distribution)
        self.set_from_dict_of_arrays(dict_of_arrays)

    def set_from_limits(self, N=10000, pos_lim=[[0, 0], [100, 100]], rad_lim=[0.1, 0.4]):
        '''
        Setup using uniform/loguniform random numbers
        '''
        # radius = np.random.uniform(rad_lim[0], rad_lim[1], size=(1, N))
        # from scipy.stats import loguniform
        # radius = loguniform.rvs(rad_lim[0], rad_lim[1], size=(1, N))
        exp = np.random.exponential(0.5, size=(1, N))
        exp = np.clip(exp, 0, 3)
        width = np.interp(exp, [0, 6], rad_lim)

        low = np.array(pos_lim)[0].reshape(2, 1)
        high = np.array(pos_lim)[1].reshape(2, 1)
        position = np.random.uniform(low, high, size=(2, N))

        yaw_deg = np.random.uniform(low=0, high=360, size=(1, N))
        pitch_deg = np.random.uniform(low=0, high=30, size=(1, N))
        aspect = np.random.uniform(low=0.5, high=1.5, size=(1, N))

        dict_of_arrays = {
            'position': position,
            'width': width,
            'height': width/2,
            'yaw_deg': yaw_deg,
            'pitch_deg': pitch_deg,
            'aspect': aspect,
        }
        self.set_from_dict_of_arrays(dict_of_arrays)

    @classmethod
    def from_arrays(cls, dict_of_arrays):
        obstacles_obj = cls()
        obstacles_obj.set_from_dict_of_arrays(dict_of_arrays)
        return obstacles_obj

    @classmethod
    def from_Y(cls, *args, **kwargs):
        ''' Return Obstacles given a terrain class '''
        obstacles_obj = cls()
        obstacles_obj.set_from_Y(*args, **kwargs)
        return obstacles_obj

    @classmethod
    def from_numpy(cls, filename):
        with open(filename, 'rb') as f:
            dict_of_arrays = np.load(f, allow_pickle=True)
            return cls.from_arrays(dict_of_arrays)

    @classmethod
    def from_yaml(cls, filename):
        with open(filename, 'r') as f:
            dict_of_lists = yaml.safe_load(f)

        # Convert lists back to NumPy arrays with the expected shapes
        dict_of_arrays = {
            # transpose back to shape (2, N)
            'position': np.array(dict_of_lists['position']).T,
            'width': np.array(dict_of_lists['width']).reshape(1, -1),
            'height': np.array(dict_of_lists['height']).reshape(1, -1),
            'yaw_deg': np.array(dict_of_lists['yaw_deg']).reshape(1, -1),
            'aspect': np.array(dict_of_lists['aspect']).reshape(1, -1),
            'pitch_deg': np.array(dict_of_lists['pitch_deg']).reshape(1, -1),
        }

        return cls.from_arrays(dict_of_arrays)

    @classmethod
    def from_limits(cls, *args, **kwargs):
        ''' Return Obstacles given a terrain class '''
        obstacles_obj = cls()
        obstacles_obj.set_from_limits(*args, **kwargs)

        return obstacles_obj

    def save_numpy(self, filename, append=False):
        dict_of_arrays = {
            'position': self.position,
            'width': self.width,
            'height': self.height,
            'yaw_deg': self.yaw_deg,
            'aspect': self.aspect,
            'pitch_deg': self.pitch_deg,
        }

        np.savez(filename, **dict_of_arrays)

    def save_yaml(self, filename):
        # Convert numpy arrays to lists so they can be serialized to YAML
        dict_of_lists = {
            'position': self.position.T.tolist(),  # transpose back to original shape
            'width': self.width.flatten().tolist(),
            'height': self.height.flatten().tolist(),
            'yaw_deg': self.yaw_deg.flatten().tolist(),
            'aspect': self.aspect.flatten().tolist(),
            'pitch_deg': self.pitch_deg.flatten().tolist(),
        }
        with open(filename, 'w') as f:
            yaml.dump(dict_of_lists, f, sort_keys=False)

    def save_pickle(self, filename, append=False):
        raise NotImplementedError

    @classmethod
    def from_pickle(cls, filename):
        raise NotImplementedError

    def translate(self, dx=0, dy=0, inplace=False):
        if inplace:
            self.position += np.array([[dx, dy]]).reshape(2, 1)
            return self
        else:
            new_obj = self.copy()
            new_obj.position += np.array([[dx, dy]]).reshape(2, 1)
            return new_obj

    def copy(self):
        new_obj = type(self)()
        new_obj.position = self.position.copy()
        new_obj.width = self.width.copy()
        new_obj.height = self.height.copy()
        new_obj.yaw_deg = self.yaw_deg.copy()
        new_obj.aspect = self.aspect.copy()
        new_obj.pitch_deg = self.pitch_deg.copy()

        return new_obj



from modules.module import Module, debug_decorator
import numpy as np


class Basic(Module):
    create_folder = False

    ''' Make basic terrains
    '''
    @debug_decorator
    def __call__(self, terrain_temp=None, scale_list=[400, 32, 0.5],
                 default=None, seed=None, **kwargs):
        from utils.noise import get_simplex
        from utils.terrains import Terrain

        terrain_temp = [] if terrain_temp is None else terrain_temp

        # Possibly replace scaling list from 'default'
        if isinstance(default, list):
            scale_list = default
        elif isinstance(default, (float, int)):
            scale_list = [default]

        # Make sure scale_list is a list
        if isinstance(scale_list, (float, int)):
            scale_list = [scale_list]

        info_dict = {}
        for i, x in enumerate(scale_list):
            scaling = 1.75/(1+x/68)
            simplex_noise = 1/scaling * get_simplex(
                **kwargs,
                scale_x=x, scale_y=x,
                info_dict=info_dict,
                logger_fn=self.info,
                seed=seed,
            )

            if seed is not None:
                seed += 1

            terrain = Terrain.from_array(simplex_noise, **info_dict)
            terrain_temp.append(terrain)

        return {
            'terrain_temp': terrain_temp,
            }


class Octaves(Module):
    create_folder = False
    ''' Make terrain by combining octaves

    Args:
      num_octaves: (int) number of octaves/terrains to generate
      start (float): start scale
      amplitude_start (float): amplitude for start octave
      persistence (float): factor by which amplitude is scaled for each octave
    '''
    @debug_decorator
    def __call__(
            self, terrain_temp=None, default=None,
            num_octaves=10,
            start=128,
            persistence=0.60,
            amplitude_start=10,
            random_amp=0.5,
            random_sign=True,
            only_generate_weights=False,
            **kwargs
    ):
        from utils.noise import get_simplex
        from utils.terrains import Terrain

        terrain_temp = [] if terrain_temp is None else terrain_temp

        # Get number of octaves from possible default input
        num_octaves = default if default is not None else num_octaves

        # Generate scale list
        start_log2 = np.log2(start)
        end_log2 = start_log2+1 - num_octaves
        scale_list = np.logspace(start_log2, end_log2, num_octaves, base=2)

        # Generate amplitude list
        amplitude_start_log = np.log(amplitude_start)/np.log(persistence)
        amplitude_end_log = amplitude_start_log+num_octaves-1
        amplitude_list = np.logspace(amplitude_start_log, amplitude_end_log,
                                     num_octaves, base=persistence)

        # Possibly multiply amplitudes by random perturbations
        if random_amp:
            random_factor = np.random.normal(
                loc=1.0, scale=random_amp, size=amplitude_list.shape)
            amplitude_list *= random_factor
        if random_sign:
            random_sign = np.random.choice([-1, 1], size=amplitude_list.shape)
            amplitude_list *= random_sign

        if not only_generate_weights:
            info_dict = {}
            for i, x in enumerate(scale_list):
                scaling = 1.75/(1+x/68)
                simplex_noise = 1/scaling * get_simplex(
                    **kwargs,
                    scale_x=x, scale_y=x,
                    center=True, info_dict=info_dict,
                    logger_fn=self.info,
                )

                terrain = Terrain.from_array(simplex_noise, **info_dict)
                terrain_temp.append(terrain)

        return {
            'terrain_temp': terrain_temp,
            # Pass weights, for use in WeightedSum
            'weights': amplitude_list,
            }


class Rocks(Module):
    create_folder = False

    ''' Test to make some rocks '''
    @debug_decorator
    def __call__(self,
                 rock_size=[0.5, 1, 2, 4],
                 rock_heights=None,
                 fraction=0.8,
                 default=None,
                 terrain_temp=None,
                 random_shift=False,
                 **kwargs):
        rock_size = default if default is not None else rock_size

        # Make sure 'rock-size' is a list
        if not isinstance(rock_size, list):
            rock_size = [rock_size]

        if rock_heights is None:
            rock_heights = np.divide(rock_size, 2)
            # [0.25, 0.5, 1, 2],

        terrain_temp = [] if terrain_temp is None else terrain_temp

        from utils.noise import get_simplex
        from utils.terrains import Terrain

        for i, (x, height) in enumerate(zip(rock_size, rock_heights)):
            info_dict = {}

            simplex_noise = get_simplex(
                scale_x=x, scale_y=x,
                info_dict=info_dict,
                center=True,
                logger_fn=self.info,
                random_shift=random_shift,
                **kwargs,
            )

            # max_value = np.max(simplex_noise)
            max_value = np.sqrt(3)/2

            # Pick all that is not holes/rocks
            pick = (simplex_noise < max_value*fraction)
            # Normalize, and remove all that is not holes/rocks
            simplex_noise = np.divide(
                simplex_noise - max_value*fraction, max_value*(1-fraction))
            simplex_noise[pick] = 0
            # Scale heights
            simplex_noise = simplex_noise * height

            terrain = Terrain.from_array(simplex_noise, **info_dict)
            terrain_temp.append(terrain)

        return {
            'terrain_temp': terrain_temp,
            }


class Holes(Module):
    create_folder = False

    ''' Test to make some rocks '''
    @debug_decorator
    def __call__(self,
                 size_list=[0.5, 1, 2, 4],
                 heights=None,
                 fraction=0.8,
                 default=None,
                 terrain_temp=None,
                 **kwargs):
        size_list = default if default is not None else size_list

        if heights is None:
            heights = np.divide(size_list, 4)

        terrain_temp = [] if terrain_temp is None else terrain_temp

        from utils.noise import get_simplex
        from utils.terrains import Terrain

        for i, (x, height) in enumerate(zip(size_list, heights)):
            info_dict = {}
            scaling = 1.75/(1+x*0.75)
            simplex_noise = 1/scaling * get_simplex(
                scale_x=x, scale_y=x,
                info_dict=info_dict,
                center=True,
                logger_fn=self.info,
                **kwargs,
            )

            min_value = np.min(simplex_noise)
            # Pick all that is not holes/rocks
            pick = (simplex_noise > min_value*fraction)
            # Normalize, and remove all that is not holes/rocks
            simplex_noise = np.divide(
                simplex_noise - min_value*fraction, -min_value*(1-fraction))
            simplex_noise[pick] = 0
            # Scale heights
            simplex_noise = simplex_noise * height

            terrain = Terrain.from_array(simplex_noise, **info_dict)
            terrain_temp.append(terrain)

        return {
            'terrain_temp': terrain_temp,
            }



import numpy as np


class Terrain():
    def __init__(self):
        self.i = 0

    @classmethod
    def from_numpy(cls, filename):
        ''' Load numpy terrain '''
        terrain_obj = cls()
        terrain_obj.array, info_dict = load_hf(filename)
        terrain_obj.size = info_dict['size']
        terrain_obj.extent = info_dict['extent']
        terrain_obj.kwargs = {}  # TODO: Added as a fix. But not in use I think

        return terrain_obj

    @classmethod
    def from_geotiff(
            cls, filename,
            center_in_origin: bool = True,
    ):
        ''' Load geotiff terrain '''
        # from geotiff import GeoTiff
        # # Open geotiff file
        # # crs_code:3006 --> sweref99 tm
        # geo_tiff = GeoTiff(filename, as_crs=3006)
        # zarr_array = geo_tiff.read()
        # ((E_min, N_max), (E_max, N_min)) = geo_tiff.tif_bBox_converted

        import tifffile as tiff

        with tiff.TiffFile(filename) as tf:
            array = tf.asarray()  # numpy array
            page = tf.pages[0]
            tags = {t.name: t.value for t in page.tags.values()}

            # Required GeoTIFF tags for north-up, no-rotation case
            # pixel size
            sx, sy, *_ = tags["ModelPixelScaleTag"]
            # (i,j,k,X,Y,Z) (may contain multiple of 6)
            tie = tags["ModelTiepointTag"]
            i0, j0, _, X0, Y0, _ = tie[:6]

            # Image size
            width = int(tags.get("ImageWidth",  page.imagewidth))
            height = int(tags.get("ImageLength", page.imagelength))

            # Compute pixel-edge bounds (common GDAL-style for north-up)
            E_min = X0 + (0 - i0) * sx
            E_max = X0 + (width - i0) * sx
            N_max = Y0 - (0 - j0) * sy
            N_min = Y0 - (height - j0) * sy

            # Fix to my convention
            array = array.T
            array = np.flip(array, axis=1)

        # Create terrain object
        terrain_obj = cls()
        terrain_obj.array = array

        # Set size and extent
        size = (E_max - E_min, N_max - N_min)
        center = ((E_max+E_min)/2, (N_max+N_min)/2)

        if center_in_origin:
            extent = (E_min-center[0], E_max-center[0],
                      N_min-center[1], N_max-center[1])
        else:
            extent = (E_min, E_max, N_min, N_max)

        terrain_obj.extent = extent
        terrain_obj.size = size

        return terrain_obj

    @classmethod
    def from_las(cls, filename):
        ''' Generate from point cloud?? '''
        raise NotImplementedError

    @classmethod
    def from_dem(cls, filename):
        ''' Load 'digital elevation model' '''
        raise NotImplementedError

    @classmethod
    def from_array(cls, array, size=None, extent=None, **kwargs):
        ''' Use array '''
        terrain_obj = cls()
        terrain_obj.array = array.astype(np.float64)

        if extent is None:
            extent = [-size[0]/2, size[0]/2, -size[1]/2, size[1]/2]
        elif size is None:
            size = [extent[1]-extent[0], extent[3]-extent[2]]
        terrain_obj.extent = extent
        terrain_obj.size = size

        # TODO: I do nothing with the kwargs,
        # but they might containt information I want to save with the array

        # TODO: La till det här, men det lirar inte exakt med hur save funkar
        # Ska man ge kwargs save, eller här, eller båda?
        terrain_obj.kwargs = kwargs

        return terrain_obj

    def save(self, filename, **kwargs):
        '''
        Save
        '''
        if not filename.endswith('.npz'):
            filename += '.npz'

        size_array = np.array(self.size).astype(np.float64)
        try:
            extent = self.extent
        except AttributeError:
            extent = [-self.size[0]/2, self.size[0]/2,
                      -self.size[1]/2, self.size[1]/2]

        with open(filename, 'wb') as f:
            np.savez(f, size=size_array, extent=extent, heights=self.array,
                     **self.kwargs, **kwargs)

    @property
    def position(self):
        '''Return the (x, y) center of the terrain from its extent'''
        x_min, x_max, y_min, y_max = self.extent
        return [(x_min + x_max) / 2, (y_min + y_max) / 2]

    @property
    def resolution(self):
        """
        Return the resolution (dx, dy) of the terrain.

        dx = (x_max - x_min) / (width - 1)
        dy = (y_max - y_min) / (height - 1)
        """
        x_min, x_max, y_min, y_max = self.extent
        height, width = self.array.shape
        dx = (x_max - x_min) / (width - 0)
        dy = (y_max - y_min) / (height - 0)

        return (dx, dy)

    # TODO: Use this instead of setting self.size etc.
    # @property
    # def size(self):
    #     '''Return (width, height) of the terrain from extent'''
    #     x_min, x_max, y_min, y_max = self.extent
    #     return [x_max - x_min, y_max - y_min]

    def __repr__(self):
        return f'Terrain(array={self.array.shape}, size={self.size}, extent={self.extent})'

    def __array__(self):
        # https://stackoverflow.com/questions/43493256/python-how-to-implement-a-custom-class-compatible-with-numpy-functions
        return self.array

    def __array_wrap__(self, obj, context=None):
        print("############### WRAP ############")
        # Make sure to return terrain object
        # Works with np 'ufunc's
        # obj is the numpy array.
        # np.maximum, but not np.maximum.reduce
        # np.add, but not np.sum1
        #

        return Terrain.from_array(obj, self.size, self.extent)


def load_hf(filename):
    '''
    Load height field from numpy binary file
    Returns an array of heights and an array specifying the size
    '''
    with open(filename, 'rb') as f:
        try:
            loaded_data = np.load(f, allow_pickle=True)
        except AttributeError:
            pass
        info_dict = {}
        for key in loaded_data.files:
            if key == 'heights':
                array = loaded_data['heights']
            else:
                info_dict[key] = loaded_data[key]

        return array, info_dict


def extent_to_size(extent):
    return [extent[1]-extent[0], extent[3]-extent[2]]


def extent_to_location(extent):
    return [(extent[1]+extent[0])/2, (extent[3]+extent[2])/2]


def merge_terrain_blocks(index_to_terrain_list):
    """
    Merges terrain arrays from a dictionary of indexed terrain lists into larger terrain arrays.

    Args:
        index_to_terrain_list (dict): A dictionary where keys are tuple indices (i, j) and 
                                      values are lists of Terrain objects at those indices.

    Returns:
        list: A list of merged Terrain objects, each representing a layer of the input terrain arrays.
    """
    # Determine the number of terrain layers and grid dimensions
    num_layers = len(next(iter(index_to_terrain_list.values())))
    last_index = list(index_to_terrain_list.keys())[-1]
    grid_rows = last_index[0] + 1
    grid_cols = last_index[1] + 1

    # Determine new extent of the merged terrain
    extent_start = list(index_to_terrain_list.values())[0][0].extent
    extent_end = list(index_to_terrain_list.values())[-1][0].extent
    merged_extent = [extent_start[0], extent_end[1], extent_start[2], extent_end[3]]

    # Prepare a 3D list to store terrain arrays
    terrain_layers = [[[None for _ in range(grid_cols)] for _ in range(grid_rows)] for _ in range(num_layers)]

    # Populate the 3D list with terrain arrays
    for (row, col), terrain_list in index_to_terrain_list.items():
        for layer_idx, terrain in enumerate(terrain_list):
            terrain_layers[layer_idx][row][col] = terrain.array

    # Convert the arrays to the desired format
    merged_terrain_list = []
    for layer_arrays in terrain_layers:
        # Merge blocks into a single array
        merged_array = np.block(layer_arrays)
        # Create a Terrain object
        merged_terrain = Terrain.from_array(merged_array, extent=merged_extent)
        merged_terrain_list.append(merged_terrain)

    return merged_terrain_list


def split_terrain_blocks(terrain_list, num_splits_row, num_splits_col):
    """
    Splits a given array into sub-arrays over two dimensions.

    Args:
        array (np.ndarray): The input array to be split into blocks.
        num_splits_row (int): The number of splits along the row dimension.
        num_splits_col (int): The number of splits along the column dimension.

    Returns:
        list: A list of lists, where each inner list contains sub-arrays split along the column dimension.
    """
    list_of_blocks = []

    for terrain in terrain_list:
        array = terrain.array
        # Split the array into sub-arrays along the row dimension
        row_splits = np.array_split(array, num_splits_row, axis=0)
        # Split each of the row sub-arrays into sub-arrays along the column dimension
        blocks = [np.array_split(row_split, num_splits_col, axis=1) for row_split in row_splits]
        list_of_blocks.append(blocks)

    blocks = np.array(list_of_blocks)

    return blocks


def assign_grid_indices(terrains):
    centers = [t.position for t in terrains]
    x_coords = sorted(set(x for x, _ in centers))
    y_coords = sorted(set(y for _, y in centers), reverse=False)

    x_map = {x: i for i, x in enumerate(x_coords)}
    y_map = {y: j for j, y in enumerate(y_coords)}

    grid_indices = [(x_map[x], y_map[y]) for x, y in centers]

    return grid_indices


def distances_to_extent(points, extent):
    """
    Calculate the shortest distance from each point to a rectangular extent.

    Parameters:
    - points: (N, 2) array of (x, y) coordinates
    - xmin, xmax: float, horizontal bounds of the extent
    - ymin, ymax: float, vertical bounds of the extent

    Returns:
    - distances: (N,) array of Euclidean distances to the rectangle
    """
    xmin, xmax, ymin, ymax = extent

    x = points[:, 0]
    y = points[:, 1]

    dx = np.where(x < xmin, xmin - x, np.where(x > xmax, x - xmax, 0))
    dy = np.where(y < ymin, ymin - y, np.where(y > ymax, y - ymax, 0))

    return np.sqrt(dx**2 + dy**2)


def get_surface_normal(hf_array, size):
    '''
    Get surface normal from array

    Args:
      hf_array:
      size: physical size (float, float)

    Return: array with surface normals
    '''
    # https://stackoverflow.com/questions/53350391/surface-normal-calculation-from-depth-map-in-python
    spacing = np.divide(size, hf_array.shape)

    zy, zx = np.gradient(hf_array, *spacing)
    normal = np.dstack((-zx, -zy, np.ones_like(hf_array)))
    n = np.linalg.norm(normal, axis=2)
    normal[:, :, 0] /= n
    normal[:, :, 1] /= n
    normal[:, :, 2] /= n

    return normal


def get_surface_area(heightmap: np.ndarray, resolution: tuple[float, float]) -> float:
    """
    Estimate the surface area of a terrain heightmap.

    Parameters:
    - heightmap: 2D numpy array of shape (H, W), representing elevation values.
    - resolution: tuple (dy, dx), the spacing between pixels in y and x directions (in meters).

    Returns:
    - surface_area: float, estimated 3D surface area in square meters.
    """
    dy, dx = resolution
    dz_dy, dz_dx = np.gradient(heightmap, dy, dx)

    # Local area element from the surface normal magnitude approximation
    area_elements = np.sqrt(1 + dz_dx**2 + dz_dy**2) * dx * dy

    return np.sum(area_elements)



from modules.module import Module, debug_decorator
import os
import sys
import numpy as np
from utils.utils import get_terrains, get_terrain


def fix_blender_path():
    try:
        import bpy
        # Make sure current dir is in path. A blender py workaround
        dir = os.path.dirname(bpy.data.filepath)
        if dir not in sys.path:
            sys.path.append(dir)
    except Exception:
        pass


def fix_blender_argv():
    # Fix to read command line arguments (after --)
    argv = sys.argv
    try:
        index = argv.index("--") + 1
    except ValueError:
        index = len(argv)
    argv = argv[index:]
    sys.argv = ['blender.py'] + argv


class Ground(Module):
    ''' Create ground in blender '''
    create_folder = False

    @debug_decorator
    def __call__(self, filename=None, default=None,
                 terrain_temp=[], terrain_prim=[],
                 ground_material=None, grid=True,
                 last=None,
                 **_):
        from utils.Blender import grid_from_array
        import hashlib

        filename = default if default is not None else filename
        if filename is not None:
            from utils.terrains import Terrain
            # Load data
            terrain = Terrain.from_numpy(filename)
            terrains = [terrain]
        else:
            # Get latest from dict/heap
            # terrain = get_terrain(terrain_temp, terrain_prim, remove=False)
            terrains = get_terrains(
                terrain_temp, terrain_prim, last, remove=False,
                print_fn=self.info)

        # Make grid from array
        for i, terrain in enumerate(terrains):
            # Generate a short hash from the terrain array
            array_bytes = terrain.array.tobytes()
            hash_digest = hashlib.sha1(array_bytes).hexdigest()[:6]  # Short hash
            name = f"Ground{i:05d}{self.file_id}_{hash_digest}"

            grid_obj = grid_from_array(
                terrain.array, name=name, size=terrain.size,
                center=terrain.position)

            if ground_material is not None:
                # Add grid image
                if grid:
                    from utils.Blender import add_grid

                    min_resolution = 2048
                    # Resolution proportional to terrain size (10 px per unit)
                    res_from_terrain = int(terrain.size[0] * 10), int(terrain.size[1] * 10)

                    # Final resolution = take maximum in each dimension
                    resolution = (max(min_resolution, res_from_terrain[0]),
                                  max(min_resolution, res_from_terrain[1]))

                    grid_kwargs = {
                        'extent': terrain.extent,
                        'size': resolution,
                        # 'filename': 'test.png',
                    }
                    add_grid(ground_material, grid_kwargs=grid_kwargs)
                grid_obj.data.materials.append(ground_material)

        # This line has not been adopted for possibly multiple objs.
        return {'grid_obj': grid_obj}


class Ground2(Module):
    """Create ground in Blender using a Displace modifier on a subdivided mesh.

    This is faster than Ground for large terrains.
    The resolution can be enhanced in blender, by modifying the subdivide

    Note that for large terrains, the grid-generation can be slow
    But this can be turned of with the 'Set:ground_material=None' module
    """
    create_folder = False

    @debug_decorator
    def __call__(self, filename=None, default=None,
                 terrain_temp=[], terrain_prim=[],
                 ground_material=None, grid=True,
                 last=None,
                 **_):

        import hashlib
        from utils.Blender import create_displaced_ground_from_terrain

        filename = default if default is not None else filename
        if filename is not None:
            from utils.terrains import Terrain
            # Load data
            terrain = Terrain.from_numpy(filename)
            terrains = [terrain]
        else:
            # Get latest from lists
            terrains = get_terrains(
                terrain_temp, terrain_prim, last, remove=False,
                print_fn=self.info)

        # Make grid from array
        for i, terrain in enumerate(terrains):
            # Generate a short hash from the terrain array
            array_bytes = terrain.array.tobytes()
            hash_digest = hashlib.sha1(array_bytes).hexdigest()[:6]  # Short hash
            name = f"Ground2{i:05d}{self.file_id}_{hash_digest}"

            grid_obj = create_displaced_ground_from_terrain(
                terrain=terrain,
                name=name,
            )

            if ground_material is not None:
                # Add grid image
                if grid:
                    from utils.Blender import add_grid

                    min_resolution = 2048
                    # Resolution proportional to terrain size (10 px per unit)
                    res_from_terrain = int(terrain.size[0] * 10), int(terrain.size[1] * 10)

                    # Final resolution = take maximum in each dimension
                    resolution = (max(min_resolution, res_from_terrain[0]),
                                  max(min_resolution, res_from_terrain[1]))

                    grid_kwargs = {
                        'extent': terrain.extent,
                        'size': resolution,
                        # 'filename': 'test.png',
                    }
                    add_grid(ground_material, grid_kwargs=grid_kwargs)
                grid_obj.data.materials.append(ground_material)

        # This line has not been adopted for possibly multiple objs.
        return {'grid_obj': grid_obj}


class BasicSetup(Module):
    '''
    Basic setup of blender scene. Remove cube and add hrdi lights
    '''
    create_folder = False

    @debug_decorator
    def __call__(self, default=None, **_):
        import bpy
        from utils.Blender import (
            get_material, setup_z_coord_shader, setup_lights,
            setup_near_far,
        )
        # Remove Cube
        from utils.Blender import remove_object
        remove_object(name='Cube')

        # Create Groundmaterial
        ground_material = get_material(name='GroundMaterial')
        ground_material.use_nodes = True

        # Setup groundmaterial
        setup_z_coord_shader(ground_material)

        # Setup lights
        sun_position = (0, 70, 70)

        use_hdri = True
        light_kwargs = {'level': 0}
        setup_lights(use_hdri, sun_position=sun_position, **light_kwargs)

        # Set standard view transform to not distort colors
        bpy.context.scene.view_settings.view_transform = 'Standard'

        setup_near_far()

        return {
            'ground_material': ground_material,
            }


class ClearScene(Module):
    """
    Removes all objects from the current Blender scene.
    """
    create_folder = False

    @debug_decorator
    def __call__(self, **_):
        import bpy

        # Deselect all objects
        bpy.ops.object.select_all(action='DESELECT')

        # Find and delete objects with names starting with "Ground"
        for obj in bpy.data.objects:
            if obj.name.startswith("Ground"):
                obj.select_set(True)

        # Delete selected objects
        bpy.ops.object.delete()

        # Optionally clear unused data blocks (meshes, materials, etc.)
        # This helps free memory and avoid clutter
        # for data_collection in [bpy.data.meshes, bpy.data.materials, bpy.data.images,
        #                         bpy.data.cameras, bpy.data.lights, bpy.data.curves]:
        for data_collection in [bpy.data.meshes]:
            for datablock in data_collection:
                if datablock.users == 0:
                    data_collection.remove(datablock)

        return {'cleared': True}


class ImageTexture(Module):
    ''' Add image texture to ground'''
    create_folder = False

    @debug_decorator
    def __call__(self, ground_material=None, texture=None, default=None, **_):
        from utils.Blender import (use_image_texture)

        texture = default if default is not None else texture
        image_texture_kwargs = {}
        use_image_texture(ground_material, texture, **image_texture_kwargs)


class ColorMap(Module):
    ''' Add image colormap'''
    create_folder = False

    @debug_decorator
    def __call__(self, ground_material=None, cmap_name='viridis', default=None, **_):
        from utils.Blender import (colormap_to_colorramp)
        import colorcet as cc
        import matplotlib.cm as cm
        cmap_name = default if default is not None else cmap_name

        if default is not None:
            # Use colorcet as default
            try:
                cmap = cc.cm[cmap_name]
            except KeyError:
                cmap = cm.get_cmap(cmap_name)

        try:
            # blender 4.x
            cr = ground_material.node_tree.nodes['Color Ramp'].color_ramp
        except KeyError:
            # blender 3.x
            cr = ground_material.node_tree.nodes['ColorRamp'].color_ramp

        colormap_to_colorramp(cr, cmap)


class GenericCamera(Module):
    ''' Setup camera'''
    create_folder = False

    @debug_decorator
    def __call__(self, camera='top', default=None, **_):
        from utils.Blender import (
            setup_camera, set_view_to_camera)

        # Setup generic camera and return
        setup_camera(resolution_x=500, resolution_y=500)
        set_view_to_camera()
        hf_info_dict = {
            'size': [50, 50],
            }

        return {
            'hf_info_dict': hf_info_dict,
        }


class Camera(Module):
    ''' Setup camera'''
    create_folder = False

    @debug_decorator
    def __call__(self, camera='top', default=None,
                 terrain_temp=[], terrain_prim=[],
                 size=None,
                 resolution=None,
                 extent=None,
                 grid_size=None,
                 angle=45,
                 center=None,
                 distance=None,
                 **_):
        from utils.artificial_shapes import determine_extent_and_resolution

        extent, (N_x, N_y) = determine_extent_and_resolution(
            resolution, size, grid_size, extent)

        from utils.Blender import (
            setup_top_camera, set_view_to_camera, setup_angled_camera)

        if isinstance(default, str):
            camera = default if default is not None else camera
            assert camera in ['top', 'angled']
        elif isinstance(default, (int, float)):
            camera = 'angled'
            angle = default

        # Get latest terrain from dict/heap
        terrain = get_terrain(terrain_temp, terrain_prim, remove=False,
                              print_fn=self.info)

        size = terrain.size
        camera_kwargs = {}
        # Setup camera
        if camera == 'top':
            camera = setup_top_camera(terrain.array, (N_x, N_y), size)
        elif camera == 'angled':
            if center is None:
                middle_z = terrain.array[int(terrain.array.shape[0]/2), int(terrain.array.shape[1]/2)]
                center = [0, 0, middle_z]
            if distance is None:
                max_size = np.max(size)
                distance = max_size*2
            camera = setup_angled_camera(
                angle=angle,
                center=center, distance=distance, **camera_kwargs)
        set_view_to_camera()


class Depth(Module):
    ''' '''
    @debug_decorator
    def __call__(self, filename="terrain.npz", default=None,
                 folder='Depth',
                 terrain_temp=[], terrain_prim=[],
                 overwrite=False,
                 **_):
        from utils.Blender import get_depth, render_eevee
        from utils.Blender import setup_render_z

        # Possibly set folder from 'default'.
        folder = default if default is not None else folder
        # Use folder
        basename = os.path.basename(self.save_dir)
        if basename != folder:
            dirname = os.path.dirname(self.save_dir)
            self.save_dir = os.path.join(dirname, folder)

        # Create folder if needed
        if not os.path.isdir(self.save_dir):
            os.makedirs(self.save_dir)

        # Get latest terrain from dict/heap
        terrain = get_terrain(terrain_temp, terrain_prim,
                              print_fn=self.info)

        setup_render_z()

        # (The render must not be done in order to retrieve the array)
        render_eevee(save_dir=self.save_dir, filename='test.png')
        dmap = -get_depth()

        # Set lowest point at 0
        dmap = dmap - np.min(dmap)

        from utils.terrains import Terrain
        terrain = Terrain.from_array(
            dmap,
            size=terrain.size,  # TODO: take from camera instead
            extent=terrain.extent,  # TODO: take from camera instead
        )

        # TODO: Fix this. Unsure how to handle this when I remove
        # the 'terrain' attribute
        terrain_prim = [terrain]

        return {
            'terrain_prim': terrain_prim,
        }


class Render(Module):
    ''' '''
    create_folder = False

    @debug_decorator
    def __call__(self, filename=None, default=None,
                 folder='Render',
                 overwrite=False,
                 loop_id=None,
                 **_):
        from utils.Blender import render_eevee

        if filename is None:
            filename = f'render{loop_id}.png'

        # Possibly set folder from 'default'.
        if default is not None and '.' in default:
            filename = default
        elif default is not None:
            folder = default

        # Use folder
        basename = os.path.basename(self.save_dir)
        if basename != folder:
            dirname = os.path.dirname(self.save_dir)
            self.save_dir = os.path.join(dirname, folder)

        # Create folder if needed
        if not os.path.isdir(self.save_dir):
            os.makedirs(self.save_dir)

        # Render
        render_eevee(save_dir=self.save_dir, filename=filename)


class RenderSegmentation(Module):
    '''
    Uniform color for 'Rock's and 'Ground'. Render and get array
    '''
    @debug_decorator
    def __call__(self, filename='render.png', default=None,
                 overwrite=False, folder='RenderSegmentation',
                 **_):
        from utils.Blender import render_eevee, get_depth
        from utils.Blender import setup_segmentation_render

        # Possibly set folder from 'default'.
        if default is not None and '.' in default:
            filename = default
        elif default is not None:
            folder = default

        # Use folder
        basename = os.path.basename(self.save_dir)
        if basename != folder:
            dirname = os.path.dirname(self.save_dir)
            self.save_dir = os.path.join(dirname, folder)

        # Create folder if needed
        if not os.path.isdir(self.save_dir):
            os.makedirs(self.save_dir)

        setup_segmentation_render()

        # Render. This can be skipped
        render_eevee(save_dir=self.save_dir, filename=filename)

        # We can use the 'get_depth' function to get the
        # 'segmentation-image' from the 'Viewer Node'
        array = get_depth()

        return {'segmentation_array': array}


class Holdout(Module):
    ''' '''
    create_folder = False

    @debug_decorator
    def __call__(self,
                 terrain_temp=[], terrain_prim=[],
                 default=None,
                 **_):
        import bpy
        bpy.context.scene.render.film_transparent = True


class AddMeshObjects(Module):
    ''' '''
    create_folder = False

    @debug_decorator
    def __call__(self,
                 filename='obstacles.npz',
                 position=[[-10, 10], [10, 10], [10, -10], [-10, -10]],
                 width=[10, 5, 3, 1],
                 yaw_deg=[0, 90, 180, 270],
                 pitch_deg=[0, 10, 20, 30],
                 terrain_temp=[], terrain_prim=[],
                 size=None,
                 default=None,
                 **_):
        # Get latest terrain from dict/heap
        terrain = get_terrain(terrain_temp, terrain_prim, remove=False,
                              print_fn=self.info)

        # Construct 'interpolator'
        if terrain is not None:
            from utils.interpolator import Interpolator
            interpolator = Interpolator(terrain.array, terrain.extent)
        else:
            interpolator = None

        import bpy
        self.position = position
        self.width = width
        self.yaw_deg = yaw_deg
        self.pitch_deg = pitch_deg

        for position, width, yaw_deg, pitch_deg in zip(
                self.position, self.width, self.yaw_deg, self.pitch_deg):
            if width > 0.29:
                print(f"position:{position}")
                rocks = [
                    'assets/Rocks/agx_data_models/convex_rock2.obj',
                    # 'assets/Rocks/agx_data_models/convex_rock3.obj',
                    # 'assets/Rocks/agx_data_models/convex_rock4.obj',
                    # 'assets/Rocks/agx_data_models/convex_rock5.obj',
                    # 'assets/Rocks/agx_data_models/convex_rock6.obj',
                ]

                import random
                obj_path = random.choice(rocks)
                # Import the OBJ file
                try:
                    # blender 4.x
                    bpy.ops.wm.obj_import(filepath=obj_path)
                except AttributeError:
                    #  blender 3.0
                    bpy.ops.import_scene.obj(filepath=obj_path)

                obj = bpy.context.selected_objects[0]

                # Set the origin to the center of mass
                bpy.context.view_layer.objects.active = obj
                bpy.ops.object.origin_set(type='ORIGIN_CENTER_OF_MASS')

                scale_factor = 0.001 * width
                bpy.ops.transform.resize(value=(
                    scale_factor, scale_factor, scale_factor))

                # TODO: size should be a tuple!!
                # x, y = np.subtract(position, size/2)
                x, y = position
                if interpolator is not None:
                    z = interpolator(x, y)
                else:
                    z = 0
                obj.location = (x, y, z)
                # obj.rotation_euler = np.random.random(3)*np.pi*2
                obj.rotation_euler = [np.deg2rad(pitch_deg), 0, np.deg2rad(yaw_deg)]
                obj.name = 'Rock'



import numpy as np
import matplotlib.pyplot as plt


# Define two angles
theta_deg = np.arange(0, 90)
theta = theta_deg*np.pi/180

alpha_deg = np.atan(2 * np.tan(theta)) * 180/np.pi

fig, ax = plt.subplots(figsize=(8, 4))
ax.plot(alpha_deg)

ax.plot(theta_deg)
ax.plot(2*theta_deg)


# ax.set_title("Histogram of Mean Slope (degrees)")
ax.set_xlabel("x-value (a.u)")
ax.set_ylabel("Mean slope degree")
ax.grid()
ax.set_ylim([0, 90])

fig.tight_layout()

# Save the figure
save_path = "test.png"
fig.savefig(save_path)
plt.close(fig)




from modules.module import Module, debug_decorator
import os
import numpy as np
import yaml
from utils.plots import save_all_axes


class Save(Module):
    create_folder = False

    ''' Save numpy terrain '''
    @debug_decorator
    def __call__(self, folder='Save',
                 default=None,
                 terrain_temp=[], terrain_prim=[],
                 overwrite=False,
                 loop_id="",
                 **_):
        # Possibly set folder from 'default'.
        folder = default if default is not None else folder

        # Use folder
        self.save_dir = os.path.join(self.save_dir_original, folder)

        list_of_filenames = []

        # Create folder if needed
        if not os.path.isdir(self.save_dir):
            os.makedirs(self.save_dir)

        # Save terrain
        for i, terrain in enumerate(terrain_prim):
            filename = f'terrain{loop_id}_{i:05d}.npz'
            filename = os.path.join(self.save_dir, filename)
            list_of_filenames.append(filename)
            if not os.path.exists(filename) or overwrite:
                terrain.save(filename)

        # Save from terrain
        for i, terrain in enumerate(terrain_temp):
            filename = f'terrain_temp{loop_id}_{i:05d}.npz'
            filename = os.path.join(self.save_dir, filename)
            list_of_filenames.append(filename)
            if not os.path.exists(filename) or overwrite:
                terrain.save(filename)

        return {
            'save_filenames': list_of_filenames
            }


class LoadTIF(Module):
    ''' Test, this should be included in Load '''
    create_folder = False

    @debug_decorator
    def __call__(self, filename=None, default=None,
                 terrain_temp=None, terrain_prim=None,
                 loop_id="", filter_on_loop_id=True,
                 **_):
        filename = default if default is not None else filename

        # Populate primary and temporary lists
        terrain_temp = [] if terrain_temp is None else terrain_temp
        terrain_prim = [] if terrain_prim is None else terrain_prim

        from utils.terrains import Terrain
        terrain = Terrain.from_geotiff(filename)
        terrain_temp.append(terrain)

        return {
            'terrain_temp': terrain_temp,
            'terrain_prim': terrain_prim,
        }


class Load(Module):
    ''' Load terrains '''
    create_folder = False

    @debug_decorator
    def __call__(self, file_list=None, default=None,
                 terrain_temp=None, terrain_prim=None,
                 loop_id="", filter_on_loop_id=True,
                 **_):
        """
        Load terrain files into structured lists.

        Args:
            file_list (list): List of filenames or patterns to load.
            default (list): Default list of filenames if file_list is None.

        # One specific terrain --> use that
        Load:runs/data_051_new_load_save/Save/terrain_temp_00_00000.npz

        # List of specific terrains --> use them
        Load:"['runs/data_051_new_load_save/Save/terrain_temp_00_00000.npz']"

        # One specific folder --> find all npz in it
        Load:runs/data_051_new_load_save/Save

        # List folders --> find npz in all of them
        Load:"['runs/data_051_new_load_save/Save']"

        # One folder, in a loop --> Load
        Loop:2 Load:runs/data_051_new_load_save/Save

        # Wildcards -->

        Returns:
            list: A list of pipes containing loaded terrain data.
        """
        # Get file-list
        file_list = default if default is not None else file_list

        # Populate primary and temporary lists
        terrain_temp = [] if terrain_temp is None else terrain_temp
        terrain_prim = [] if terrain_prim is None else terrain_prim

        # Parse any wildcards, and make sure it is a list
        file_list = self.parse_wildcards(file_list)
        # Check that files exist, and possibly prepend 'save-dir'
        file_list = self.check_exist_and_possibly_prepend_savedir(file_list)
        # Expand folders to get npz files
        file_list = self.expand_folders_to_get_npz(file_list)

        import re
        from utils.terrains import Terrain
        # Regular expressions to match filenames

        terrain_regex = re.compile(r"terrain(?:(_.*?))?_(\d{5})\.npz$")
        terrain_temp_regex = re.compile(r"terrain_temp(?:(_.*?))?_(\d{5})\.npz$")

        terrain_list = self.filter_files_by_regex(
            file_list, terrain_regex, loop_id, filter_on_loop_id)
        terrain_temp_list = self.filter_files_by_regex(
            file_list, terrain_temp_regex, loop_id, filter_on_loop_id)

        # Load to create terrain
        for filename in terrain_list:
            terrain = Terrain.from_numpy(filename)
            terrain_prim.append(terrain)

        # Load to create terrain_temp
        for filename in terrain_temp_list:
            terrain = Terrain.from_numpy(filename)
            terrain_temp.append(terrain)

        return {
            'terrain_temp': terrain_temp,
            'terrain_prim': terrain_prim,
        }

    def filter_files_by_regex(self, file_list, regex, loop_id, filter_on_loop_id=True):
        """Return subset of files matching regex and (optionally) loop_id."""
        new_file_list = []

        for filename in file_list:
            base = os.path.basename(filename)
            m = regex.match(base)
            if not m:
                continue

            file_loop_id = m.group(1) or ""  # extracted loop id
            if (not filter_on_loop_id or file_loop_id.startswith(loop_id)) and 'temp' not in file_loop_id:
                new_file_list.append(filename)

        return new_file_list

    def parse_wildcards(self, file_list):
        # Draft of handling wildcards
        # Not sure how to best handle this systematically
        if '*' in file_list:
            import glob
            file_list = glob.glob(file_list)
            file_list.sort()

        if '[' in file_list:
            import glob
            file_list = glob.glob(file_list)
            file_list.sort()

        if not isinstance(file_list, list):
            file_list = [file_list]

        return file_list

    def check_exist_and_possibly_prepend_savedir(self, file_list):
        new_file_list = []
        for filename in file_list:
            # Check if filename is a file or folder
            if os.path.isfile(filename):
                new_file_list.append(filename)
            elif os.path.isdir(filename):
                new_file_list.append(filename)
            else:
                # Prepend save_dir if *that* file exists
                test = os.path.join(self.save_dir, filename)
                if os.path.exists(test):
                    new_file_list.append(os.path.join(self.save_dir, filename))
                else:
                    # Or raise error
                    raise FileNotFoundError

        return new_file_list

    def expand_folders_to_get_npz(self, file_list):
        from utils.utils import get_list_of_files
        # If folder, get all npz files in it
        new_file_list = []
        for filename in file_list:
            if os.path.isdir(filename):
                folder = filename
                files = get_list_of_files([folder], '.npz')
                for f in files:
                    new_file_list.append(os.path.join(folder, f))
            else:
                new_file_list.append(filename)

        return new_file_list


class ClearTerrain(Module):
    @debug_decorator
    def __call__(self,
                 hf_array=None, hf_info_dict=None,
                 call_number=None, call_total=None, **pipe):

        return {
            'terrain_temp': [],
            'terrain_prim': [],
        }


class CombinePipes(Module):
    ''' Combine terrains from different pipes '''
    create_folder = False

    @debug_decorator
    def __call__(self, operation='Add', default=None,
                 hf_array=None, hf_info_dict=None,
                 call_number=None, call_total=None, **pipe):

        operation = default if default is not None else operation

        # If only one call, do nothing
        if call_total == 1:
            return

        if call_number == 0:
            # first call
            self.arrays = [hf_array]
            return 'remove'
        elif call_number + 1 != call_total:
            # not last call
            self.arrays.append(hf_array)
            return 'remove'

        # on last call
        self.arrays.append(hf_array)

        if operation == 'Add':
            hf_array = np.sum(self.arrays, axis=0)
        elif operation == 'Max':
            hf_array = np.maximum.reduce(self.arrays)
        elif operation == 'Min':
            hf_array = np.minimum.reduce(self.arrays)
        elif operation == 'Prod':
            hf_array = np.prod(self.arrays, axis=0)
        else:
            raise AttributeError

        return {
            'hf_array': hf_array,
            'hf_info_dict': hf_info_dict,
        }


class FindRocks(Module):
    """
    Detect rock-like features from terrain data and return them as 'obstacles'.

    Identifies local maxima in elevation, estimates their real-world position,
    height, and width.
    """
    @debug_decorator
    def __call__(self, terrain_prim=None, terrain_temp=None,
                 default=None, last=None, **kwargs):
        from utils.utils import get_terrains, find_rocks

        terrains = get_terrains(
            terrain_temp, terrain_prim, last=last, remove=False,
            print_fn=self.info)

        # position, height, size = zip(*(find_rocks(t) for t in terrains))
        position_list = []
        height_list = []
        size_list = []
        terrain_number_list = []

        for i, terrain in enumerate(terrains):
            position, height, size = find_rocks(terrain)
            position_list.append(position)
            height_list.append(height)
            size_list.append(size)
            terrain_number_list.append(np.full(len(height), i))

        position = np.concatenate(position_list)
        height = np.concatenate(height_list)
        size = np.concatenate(size_list)
        terrain_number = np.concatenate(terrain_number_list)

        pixel_diag = np.linalg.norm(terrains[0].resolution)
        width = np.sqrt(size) * pixel_diag

        result = {
            'position': position,
            'height': height,
            'width': width,
            'terrain_number': terrain_number,
            'yaw_deg': np.zeros_like(height),
            'pitch_deg': np.zeros_like(height),
            'aspect': np.ones_like(height),
        }

        # Filter to include only rocks of some height
        pick = (height > 0.1)
        result = {k: v[pick] for k, v in result.items()}

        return result


class PlotRocks(Module):
    ''' Plot rocks
    '''
    @debug_decorator
    def __call__(self, terrain_prim=None, terrain_temp=None,
                 position=None, width=None,
                 default=None, last=None,
                 dpi=200, exportmode=False,
                 **kwargs):
        from utils.utils import get_terrain
        from utils.plots import plot_terrain, new_fig

        # Get terrain
        try:
            terrain = get_terrain(
                terrain_temp, terrain_prim, last=last, remove=False,
                print_fn=self.info)
        except AttributeError:
            terrain = None
        except TypeError:
            terrain = None

        filename = f'rocks{self.file_id}.png'
        filename = os.path.join(self.save_dir, filename)

        if terrain is not None:
            fig, ax = plot_terrain(terrain)
        else:
            fig, ax = new_fig()

        ax.scatter(
            # position[:, 0], position[:, 1],
            *position.T,
            facecolor='none', color='red', s=40*width**2)
        fig.savefig(filename, dpi=dpi)
        if exportmode:
            save_all_axes(fig, filename, delta=0.0, dpi=dpi)
            filename = filename.replace('.png', f'_{0:01d}.png')
            # Possibly use as image texture
            return {'texture': filename}


class Plot(Module):
    create_folder = False

    ''' Plot terrain '''
    @debug_decorator
    def __call__(self, default=None, overwrite=False,
                 terrain_temp=[], terrain_prim=[], folder='Plot',
                 filename_base='terrain',
                 exportmode=False, dpi=400,
                 vmin=None, vmax=None,
                 call_number=None, call_total=None, **kwargs):
        from utils.plots import plot_terrain, save_all_axes

        # Possibly set folder from 'default'.
        folder = default if default is not None else folder

        list_of_filenames = []

        # Use folder
        basename = os.path.basename(self.save_dir)
        if basename != folder:
            dirname = os.path.dirname(self.save_dir)
            self.save_dir = os.path.join(dirname, folder)

        # Create folder if needed
        if not os.path.isdir(self.save_dir):
            os.makedirs(self.save_dir)

        # Plot terrain
        for i, terrain in enumerate(terrain_prim):
            filename = f'{filename_base}{self.file_id}_{i:05d}.png'
            filename = os.path.join(self.save_dir, filename)
            list_of_filenames.append(filename)
            if not os.path.exists(filename) or overwrite:
                fig, ax = plot_terrain(terrain, vmin=vmin, vmax=vmax)
                fig.savefig(filename, dpi=dpi)
                if exportmode:
                    save_all_axes(fig, filename, delta=0.0, dpi=dpi)

        # Plot terrain_temp
        for i, terrain in enumerate(terrain_temp):
            filename = f'{filename_base}_temp{self.file_id}_{i:05d}.png'
            filename = os.path.join(self.save_dir, filename)
            list_of_filenames.append(filename)
            if not os.path.exists(filename) or overwrite:
                fig, ax = plot_terrain(terrain, vmin=vmin, vmax=vmax)
                fig.savefig(filename, dpi=dpi)
                if exportmode:
                    save_all_axes(fig, filename, delta=0.0, dpi=dpi)

        result = {'plot_filenames': list_of_filenames}

        if exportmode:
            filename = filename.replace('.png', f'_{0:01d}.png')
            # Possibly use last result as image texture
            result['texture'] = filename

        return result


class SaveYaml(Module):
    ''' Plot terrain '''
    @debug_decorator
    def __call__(self, default=None, overwrite=False,
                 call_number=None, call_total=None,
                 terrain_temp=[], terrain_prim=[],
                 **kwargs):
        # Initialize
        if call_number == 0:
            self.list_of_kwargs = []

        # Collect data
        self.list_of_kwargs.append(
            {**kwargs,
             'call_total': call_total,
             'call_number': call_number,
             })

        # Return, except on last call
        if call_number+1 != call_total:
            return

        # Save
        filename = 'data.yml'
        filename = os.path.join(self.save_dir, filename)
        with open(filename, 'w') as outfile:
            yaml.dump(
                self.list_of_kwargs, outfile,
                default_flow_style=None,
                allow_unicode=True,)



from opensimplex import noise2array
import numpy as np
import opensimplex


def get_simplex(resolution=None, size=None, N=None, grid_size=None,
                scale_x=50, scale_y=50, seed=None,
                extent=None,
                info_dict=None,
                random_shift=False,
                logger_fn=None, **_):
    '''
    Get noise with shape (Nx, Ny)

    Can be defined using any two of size, points-per-meter, and total points
    Tuple input overrides individual values

    Args:
    '''
    if seed is None:
        opensimplex.random_seed()
    else:
        opensimplex.seed(seed)

    # Setup sizes etc.
    from utils.artificial_shapes import determine_extent_and_resolution
    extent, (N_x, N_y) = determine_extent_and_resolution(
        resolution, size, grid_size, extent)

    size_x = extent[1] - extent[0]
    size_y = extent[3] - extent[2]

    if info_dict is not None:
        # Add to the info-dict if provided
        info_dict['extent'] = extent
        info_dict['opensimplex_seed'] = opensimplex.get_seed()
        info_dict['N'] = (N_x, N_y)
        info_dict['size'] = (size_x, size_y)

    if logger_fn is not None:
        info = f"points:{info_dict['N']}, " +\
            f"extent:{info_dict['extent']}, " +\
            f"seed:{info_dict['opensimplex_seed']}"
        logger_fn(info)

    if random_shift:
        # Add random shift, connected to opensimplex seed to get
        # repeatable results
        import random
        seed = opensimplex.get_seed()
        rng = random.Random(seed)
        # Generate two random numbers in [0, 1)
        rx = rng.random()
        ry = rng.random()
        # Add random shift to the extent
        import copy
        extent = copy.deepcopy(extent)
        extent[:2] = np.add(extent[:2], rx*scale_x)
        extent[2:] = np.add(extent[2:], ry*scale_y)

    x = np.linspace(extent[0]/scale_x, extent[1]/scale_x, num=N_x, endpoint=False)
    y = np.linspace(extent[2]/scale_y, extent[3]/scale_y, num=N_y, endpoint=False)

    # 'shift' indices due to noise2array definition
    return noise2array(y, x)


def generate_points_from_2D_prob(prob_2d, N=100):
    '''
    Sample N random points from a 2D probability distribution

    Return:
      np.array of shape (2, N) of dtype int64
    '''
    # Update the 'flat_probability'
    flat_probability = prob_2d.flatten()

    # Construct generator
    rng = np.random.default_rng()

    # Generate samples
    sample_indices = rng.choice(
        a=flat_probability.size,
        size=N,
        p=flat_probability,
    )

    # Unravel indices to 2D shape again
    adjusted_indices = np.unravel_index(
        sample_indices, prob_2d.shape)

    return np.array(adjusted_indices)



import numpy as np
import os
from scipy.ndimage import gaussian_filter
from scipy.optimize import brentq

from modules.module import Module, debug_decorator
from utils.utils import get_terrains, get_terrain
from utils.plots import new_fig, save_all_axes


class SaveData(Module):
    """
    Save selected list or array data to a compressed .npz file.
    """
    create_folder = False

    @debug_decorator
    def __call__(self,
                 filename='data.npz',
                 default=None,
                 folder='SaveData',
                 # parameters to skip:
                 terrain_temp=[], terrain_prim=[],
                 last=None, call_number=None,
                 call_total=None, size=None, resolution=None,
                 extent=None, loop_id=None, loop_id_r=None,
                 **kwargs):
        # use default to update filename or folder
        if default is not None and '.npz' in default:
            filename = default
        if default is not None:
            folder = default

        # Use folder
        basename = os.path.basename(self.save_dir)
        if basename != folder:
            dirname = os.path.dirname(self.save_dir)
            self.save_dir = os.path.join(dirname, folder)

        # Create folder if needed
        if not os.path.isdir(self.save_dir):
            os.makedirs(self.save_dir)

        # Create dict to collect data to save
        data = {}

        # Full path
        filename = os.path.join(self.save_dir, filename)

        for k, v in kwargs.items():
            # Only save list-like objects
            if isinstance(v, (list, tuple, np.ndarray)):
                data[k] = np.array(v)

        # Spara till .npz
        np.savez_compressed(filename, **data)
        self.info(f"Saved to {filename}")


class LoadData(Module):
    """
    Load a .npz file and return the contents as a dictionary of arrays.
    """
    create_folder = False

    @debug_decorator
    def __call__(self,
                 default=None,
                 filename=None,
                 overwrite=False,
                 **kwargs,
                 ):
        if default is not None:
            filename = default
        else:
            original_save_dir = os.path.dirname(self.save_dir)
            for name in ['SaveData/data.npz', 'LogData/data.npz']:
                filename = os.path.join(original_save_dir, name)
                if os.path.isfile(filename):
                    break

        if not os.path.isfile(filename):
            raise FileNotFoundError(f"File not found: {filename}")

        # Load file
        data = np.load(filename, allow_pickle=True)

        # Convert to dict
        result = {}
        for k, v in data.items():
            if not overwrite and k in kwargs:
                # Update name to avoid overwrite
                k = f'{k}_2'
            result[k] = v

        self.info(f"Loaded from {filename}: keys = {list(result.keys())}")

        return result


class LogData(Module):
    """
    Collect data by appending to lists, then save to .npz file on delete.
    """
    create_folder = True

    def __init__(self, *args, **kwargs):
        self.data = {}  # key: list of values
        self.filename = None
        super().__init__(*args, **kwargs)

    @debug_decorator
    def __call__(
            self, default=None, filename='data.npz',
            # parameters to skip:
            terrain_temp=[], terrain_prim=[],
            last=None,
            call_total=None, size=None,
            extent=None, loop_id=None, loop_id_r=None,
            # don't skip
            # call_number=None,
            # resolution=None,
            **kwargs,
    ):

        # Append each kwarg value to the internal data lists
        for k, v in kwargs.items():
            if k not in self.data:
                self.data[k] = []
            self.data[k].append(v)

        # Possibly set the filename
        if self.filename is None:
            if default is None:
                self.filename = filename
            if default is not None:
                self.filename = default

    def save(self):
        # Convert lists to arrays and save
        to_save = {}
        for k, v in self.data.items():
            # Only save if it's a proper NumPy array
            try:
                arr = np.array(v)
                # if arr.dtype is not object:
                if not np.issubdtype(arr.dtype, np.object_):
                    to_save[k] = arr
            except Exception:
                # Skip entries that can't be converted to a NumPy array
                pass

            # Second attempt, try to concatenate arrays/lists of different size
            if k not in to_save:
                try:
                    arr = np.concatenate(v)
                    # if arr.dtype is not object:
                    if not np.issubdtype(arr.dtype, np.object_):
                        to_save[k] = arr
                except Exception:
                    pass

        filename = os.path.join(self.save_dir, self.filename)
        np.savez(filename, **to_save)
        self.info(f"Saved log data to {filename}")

    def __del__(self):
        # Save on delete
        if self.data:
            self.save()


class MakeData1D(Module):
    """
    Split all arrays with shape (N, D) into D separate 1D arrays named
    key_0, key_1, ..., key_{D-1}.
    """
    create_folder = False

    @debug_decorator
    def __call__(
            self,
            # parameters to skip:
            terrain_temp=[], terrain_prim=[],
            last=None, call_number=None,
            call_total=None, size=None, resolution=None,
            extent=None, loop_id=None, loop_id_r=None,
            **kwargs,
    ):
        result = {}
        for k, v in kwargs.items():
            if isinstance(v, np.ndarray) and v.ndim == 2 and v.shape[1] > 1:
                for i in range(v.shape[1]):
                    result[f"{k}_{i}"] = v[:, i]
                self.info(f"Split '{k}' into {v.shape[1]} separate 1D arrays.")
            else:
                result[k] = v  # Keep unchanged if not 2D or not wide enough

        return result


class Roughness(Module):
    """
    Calculate terrain roughness by comparing the surface area of the original
    terrain to a smoothed version of the terrain.

    Roughness is computed as:
        roughness = original_surface_area / smoothed_surface_area

    Parameters:
    - sigma_meter (float): The smoothing length in meters (default: 5).
    """
    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=[], terrain_prim=[],
                 default=None, last=None, sigma_meter=5, **_):

        from utils.terrains import get_surface_area

        # Possibly set sigma (m) using default value
        sigma_meter = default if default is not None else sigma_meter

        terrains = get_terrains(
            terrain_temp, terrain_prim, last, remove=False,
            print_fn=self.info)

        results = {'roughness': []}

        for i, terrain in enumerate(terrains):
            resolution = terrain.resolution  # (dy, dx)
            sigma_grid = np.divide(sigma_meter, resolution)

            # Apply Gaussian smoothing
            smoothed_array = gaussian_filter(terrain.array, sigma=sigma_grid)

            # Compute surface normals
            area_orig = get_surface_area(terrain.array, terrain.resolution)
            area_smooth = get_surface_area(smoothed_array, terrain.resolution)

            roughness = area_orig / area_smooth if area_smooth > 0 else np.nan

            self.info(f"Terrain {i}: roughness = {roughness:.4f}")
            results['roughness'].append(roughness)

        return results


class Slope(Module):
    """
    Compute mean slope and gradient for terrain arrays.

    Parameters:
    - sigma_meter (float): Smoothing length in meters (default: 0).
    """
    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=[], terrain_prim=[],
                 default=None, last=None, sigma_meter=0, **_):

        # Possibly override sigma with `default`
        sigma_meter = default if default is not None else sigma_meter

        terrains = get_terrains(
            terrain_temp, terrain_prim, last, remove=False,
            print_fn=self.info)

        results = {
            'slope_deg': [],  # mean slope, in degrees
            'mean_gradient': [],  # mean gradient (dz/dx and dz/dy)
        }

        for i, terrain in enumerate(terrains):
            # Get terrain resolution
            resolution = terrain.resolution  # (dy, dx)

            # Calculate gradient, with or without smoothing
            if sigma_meter > 0:
                sigma_grid = np.divide(sigma_meter, resolution)
                # Apply Gaussian smoothing
                smoothed = gaussian_filter(terrain.array, sigma=sigma_grid)
                # Compute gradients in physical units
                dz_dx, dz_dy = np.gradient(smoothed, *resolution)
            else:
                # Without smoothing
                dz_dx, dz_dy = np.gradient(terrain.array, *resolution)

            # Get mean derivatives in 'metre' scale.
            mean_dz_dx = np.nanmean(dz_dx)
            mean_dz_dy = np.nanmean(dz_dy)

            mean_gradient = (mean_dz_dx, mean_dz_dy)
            norm = np.linalg.norm(mean_gradient)
            mean_slope_deg = np.atan(norm)*180/np.pi

            # # Test: Use edge-values instead
            # mean_height_at_xmin = np.mean(terrain.array[0, :])
            # mean_height_at_xmax = np.mean(terrain.array[-1, :])
            # mean_height_at_ymin = np.mean(terrain.array[:, 0])
            # mean_height_at_ymax = np.mean(terrain.array[:, -1])
            # marginal_mean_heights = [
            #     mean_height_at_xmin, mean_height_at_xmax,
            #     mean_height_at_ymin, mean_height_at_ymax]
            # # Get extent
            # [xmin, xmax, ymin, ymax] = terrain.extent
            # # Remove one cell-length in each direction, to get the same value
            # # as mean of derivative above.
            # dz_dx2 = (mean_height_at_xmax-mean_height_at_xmin)/(xmax-xmin - resolution[0])
            # dz_dy2 = (mean_height_at_ymax-mean_height_at_ymin)/(ymax-ymin - resolution[1])

            self.info(f"Terrain {i}: slope = {mean_slope_deg:.4f} (deg)")

            results['slope_deg'].append(mean_slope_deg)
            results['mean_gradient'].append(mean_gradient)

        return results


class SurfaceStructure(Module):
    """
    Compute surface-structure measures
    """
    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=[], terrain_prim=[],
                 position=None, height=None,
                 default=None, last=None, sigma_meter=0, **_):

        terrain = get_terrain(
            terrain_temp, terrain_prim, remove=False,
            print_fn=self.info)

        # Count number of rocks larger than some cutoffs
        a = np.sum(height > 0.1)
        b = np.sum(height > 0.3)
        c = np.sum(height > 0.5)
        d = np.sum(height > 0.7)
        self.debug(f"(a, b, c, d):{(a, b, c, d)}")

        # Categorize as number of obstaces per ha (100x100m)
        per_ha_factor = 100*100/np.multiply(*terrain.size)

        results = {
            'h_20': (a-b)*per_ha_factor,
            'h_40': (b-c)*per_ha_factor,
            'h_60': (c-d)*per_ha_factor,
            'h_80': (d-0)*per_ha_factor,
            'h_sum': a*per_ha_factor,
        }

        # Calculate surface structure
        from utils.utils import calc_ytstruktur_2
        h_20 = results['h_20']
        h_40 = results['h_40']
        h_60 = results['h_60']
        h_80 = results['h_80']
        Y = calc_ytstruktur_2(h_20, h_40, h_60, h_80)

        results['Y'] = Y
        self.info(f"Y:{Y}")

        return results


class CombineRoughness(Module):
    """
    Combine roughness values and weights into a proxy roughness estimate.
    """
    list_or_results = []
    create_folder = False

    @debug_decorator
    def __call__(self, roughness=None, weights=0,
                 default=None, last=None, **kwargs):
        """
        Parameters:
        - roughness: Array-like of roughness values.
        - weights: Array-like of weights for each roughness value.
        """
        r_i = roughness
        w_i = weights

        alpha = 1.0  # Currently hardcoded

        def model(alpha, beta=2):
            base = (np.power(r_i, alpha)-1) * np.power(np.abs(w_i), beta)
            r_pred = 1 + np.power(np.sum(base), 1 / alpha)
            return r_pred

        result = model(alpha)

        self.info(f"proxy roughness:{result}")
        return {'proxy_roughness': result}


class CombineSlope(Module):
    """
    Combine slope gradients into a single representative slope estimate.

    Takes weighted mean gradients (dz/dx, dz/dy) and computes the resulting
    overall slope angle in degrees.
    """
    list_or_results = []
    create_folder = False

    @debug_decorator
    def __call__(self, mean_gradient=None, weights=0,
                 default=None, last=None, **kwargs):
        """
        Parameters:
        - mean_gradient: List of (dz/dx, dz/dy) tuples.
        - weights: List or array of weights for each gradient.
        """
        combined_gradient = np.zeros(2)

        for (dz_dx, dz_dy), weight in zip(mean_gradient, weights):
            # Scale slopes by weight
            dz_dx *= weight
            dz_dy *= weight

            # Add to combined vector
            combined_gradient += np.array([dz_dx, dz_dy])

        self.info(f"combined_gradient:{combined_gradient}")
        norm = np.linalg.norm(combined_gradient)
        combined_slope_deg = np.atan(norm)*180/np.pi
        self.info(f"combined_slope_deg:{combined_slope_deg}")

        return {'combined_slope_deg': combined_slope_deg}


class OptimizeRoughness(Module):
    """
    Find the optimal roughness proxy

    """
    create_folder = True

    @debug_decorator
    def __call__(self, mean_gradient=None, weights=0,
                 default=None, last=None, **kwargs):
        from scipy.optimize import minimize_scalar

        original_save_dir = os.path.dirname(self.save_dir)

        file1 = os.path.join(original_save_dir, "LogData/data.npz")
        data1 = np.load(file1)

        file2 = os.path.join(original_save_dir, "LogData/data2.npz")
        data2 = np.load(file2)

        w_i = data1['weights']
        r_i = data1['roughness']

        r_obs = data2['roughness']

        def model(alpha, beta=2):
            # r_pred = 1 + np.sum((np.power(r_i, alpha)-1) * np.abs(w_i), axis=1)
            # beta = alpha
            base = (np.power(r_i, alpha)-1) * np.power(np.abs(w_i), beta)
            r_pred = 1 + np.power(np.sum(base, axis=1), 1 / alpha)

            # r_pred = 1.0 + np.sum((np.power(r_i-1, alpha)) * np.abs(w_i), axis=1)

            return r_pred

        # Loss function to minimize
        def loss(alpha):
            r_pred = model(alpha)
            loss = np.sum((r_obs - r_pred)**2)
            self.debug(f"(alpha, loss):{(alpha, loss)}")

            return loss

        y = []
        # Set the limits of the plots *and the optimization*
        x = np.linspace(0.55, 2, 100)

        for x_ in x:
            y.append(loss(x_))
        y = np.array(y)

        # Plot loss function as a function of alpha
        fig, ax = new_fig()
        ax.plot(x, y)
        filename = os.path.join(self.save_dir, 'loss_as_function_of_alpha.png')
        fig.savefig(filename)

        # Minimize the loss
        res = minimize_scalar(loss, bounds=(x[0], x[-1]), method='bounded')
        if res.success:
            alpha = res.x
            self.debug(f"Optimal alpha: {alpha}")
            self.debug(f"loss(alpha):{loss(alpha)}")
        else:
            logger.error("Optimization failed.")

        # Plot roughness v pred roughness
        fig, ax = new_fig()
        r_pred = model(alpha)
        ax.scatter(r_obs, r_pred)
        ax.set_xlabel('r-obs')
        ax.set_ylabel('r-pred')
        ax.grid()
        ax.plot([1, 1.15], [1, 1.15])
        filename = os.path.join(self.save_dir, 'pred_v_obs.png')
        fig.savefig(filename)


class SetRoughness(Module):
    """
    Adjust weights so that the proxy roughness matches the target.

    The proxy roughness is defined as:
        R_proxy = 1 + sum_i (sqrt(r_i) - 1) * w_i

    We solve for lambda in the following,
        R_target = 1 + sum_i (sqrt(r_i) - 1) * w_i * r_i^λ

    Where:
        - r_i is the roughness of terrain i
        - w_i is the original weight of terrain i
        - λ (lambda) is a scalar exponent we solve for
        - The adjusted weight is: w_i_target = w_i * r_i^λ

    This allows both positive and negative weights, and ensures that weights
    scale toward or away from zero depending on how much each terrain contributes
    to the roughness deviation.
    """
    create_folder = False

    @debug_decorator
    def __call__(self, roughness=None, weights=None, target_roughness=1.0,
                 default=None, **_):

        # Use `default` if `target_roughness` is not explicitly provided
        target_roughness = default if default is not None else target_roughness

        # Check input validity
        if roughness is None or weights is None:
            raise ValueError("Both roughness and weights must be provided.")

        # Convert to numpy arrays for vectorized computation
        r_i = np.asarray(roughness)
        weights = np.asarray(weights)

        # Define proxy roughness function
        def proxy_roughness(lambda_, alpha=1, beta=2):
            # Scale weights
            w_i = np.absolute(weights) * r_i**lambda_

            # Calculate proxy roughness using scaled weights
            base = (np.power(r_i, alpha)-1) * np.power(np.abs(w_i), beta)
            r_pred = 1 + np.power(np.sum(base), 1 / alpha)

            return r_pred

        # Define root-finding function: f(λ) = R(λ) - target_roughness
        def f(lambda_):
            return proxy_roughness(lambda_) - target_roughness

        # Solve for λ numerically using Brent's method
        lambda_solution = brentq(f, -1000, 100)

        self.debug(f"lambda_solution:{lambda_solution}")

        # Compute new weights: w_i_new = w_i * r_i^λ
        adjusted_weights = weights * r_i**lambda_solution
        self.debug(f"adjusted_weights:{adjusted_weights}")

        return {'weights': adjusted_weights}


class SetSlope(Module):
    """
    Adjust weights so that the proxy slope matches the target.
    """
    create_folder = False

    @debug_decorator
    def __call__(self, slope_deg=None, mean_gradient=None, weights=None,
                 target_slope_deg=5, default=None, **_):

        # Use `default` if `target_slope_deg` is not explicitly provided
        target_slope_deg = default if default is not None else target_slope_deg

        target_slope = np.tan(np.radians(target_slope_deg))

        gradients = np.asarray(mean_gradient)
        weights = np.asarray(weights)

        # Norm (magnitude) of each gradient
        norms = np.linalg.norm(gradients, axis=1)

        # Simple rescale
        norms = 1+norms/np.max(norms)

        # Define proxy slope as function of λ
        def proxy_slope(lambda_):
            scaled_weights = weights * norms**lambda_
            combined_gradient = np.sum(gradients.T * scaled_weights, axis=1)
            return np.linalg.norm(combined_gradient)

        # Root function: difference from target slope
        def f(lambda_):
            return proxy_slope(lambda_) - target_slope
        try:
            lambda_solution = brentq(f, -10, 100)
        except ValueError:
            self.warning("Brentq failed: returning original weights")
            return {'weights': weights}

        adjusted_weights = weights * norms**lambda_solution

        self.info(f"lambda_solution: {lambda_solution:.4f}")
        self.info(f"adjusted_weights: {adjusted_weights}")

        return {'weights': adjusted_weights}


class PlotScatter(Module):
    """
    Plot scatter plots of all pairs of data with equal length.
    Skips plotting when keys are identical (k1 == k2) or lengths differ.
    """
    create_folder = True

    @debug_decorator
    def __call__(self,
                 default=None,
                 # parameters to skip:
                 terrain_temp=[], terrain_prim=[],
                 last=None, call_number=None,
                 call_total=None, size=None, resolution=None,
                 extent=None, loop_id=None, loop_id_r=None,
                 # plot-parameters
                 exportmode=False, dpi=200, overwrite=False,
                 color=None, cmap=None,
                 grid=False,
                 **kwargs):

        # Loop over all key-value pairs in kwargs
        for k1, v1 in kwargs.items():
            for k2, v2 in kwargs.items():
                # Skip plotting identical variables or mismatched lengths
                # if k1 == k2 or len(v1) != len(v2):
                if k1 == k2:
                    continue

                # Generate filename and save figure
                filename = f"{k1}_v_{k2}{self.file_id}.png"
                filename = os.path.join(self.save_dir, filename)
                if os.path.isfile(filename) and not overwrite:
                    continue

                try:
                    # Create figure and axis
                    fig, ax = new_fig()

                    if color is not None and color in kwargs:
                        color_data = kwargs[color]
                        sc = ax.scatter(v1, v2, c=color_data, cmap=cmap, alpha=1)
                        fig.colorbar(sc, ax=ax, label=color)
                    else:
                        ax.scatter(v1, v2)

                    # Label axes
                    ax.set_xlabel(k1)
                    ax.set_ylabel(k2)

                    # Apply tight layout
                    fig.tight_layout()

                    if grid:
                        ax.grid()
                    fig.savefig(filename)
                    if exportmode:
                        save_all_axes(fig, filename, delta=0.0, dpi=dpi)

                except Exception:
                    # Skip silently on error (could add logging if needed)
                    pass


class PlotHistogram(Module):
    """
    Plot histogram for each entry in kwargs.
    Accepts optional `bins` and `range` for histogram computation.
    """
    create_folder = True

    @debug_decorator
    def __call__(self,
                 default=None,
                 # parameters to skip:
                 terrain_temp=[], terrain_prim=[],
                 last=None, call_number=None,
                 call_total=None, size=None, resolution=None,
                 extent=None, loop_id=None, loop_id_r=None,
                 # plot parameters
                 bins=10, range=None,
                 **kwargs):

        # Use `default` as bins if specified
        bins = default if default is not None else bins

        for k, v in kwargs.items():
            try:
                # Create figure and axis
                fig, ax = new_fig()

                # Compute histogram
                hist, bin_edges = np.histogram(v, bins, range)

                # Compute bin centers
                bin_centers = 0.5 * (bin_edges[:-1] + bin_edges[1:])

                # Plot as bar chart using bin centers
                ax.bar(bin_centers, hist, width=0.9*np.diff(bin_edges))

                # Label axes
                ax.set_xlabel(k)
                ax.set_ylabel('Occurrences')

                # Apply tight layout
                fig.tight_layout()

                # Generate filename and save figure
                filename = f"{k}{self.file_id}.png"
                filename = os.path.join(self.save_dir, filename)
                fig.savefig(filename)

            except Exception:
                # Silently skip any error (consider logging)
                pass


class PlotLines(Module):
    """
    Plot line graphs for each entry in kwargs.

    Each value `v` is assumed to be 2D (n_lines x n_points). Lines are
    plotted by transposing the array so each row becomes a line on the plot.

    The x-axis represents the index of each point.
    """
    create_folder = True

    @debug_decorator
    def __call__(self,
                 default=None,
                 # parameters to skip:
                 terrain_temp=[], terrain_prim=[],
                 last=None, call_number=None,
                 call_total=None, size=None, resolution=None,
                 extent=None, loop_id=None, loop_id_r=None,
                 **kwargs):

        for k, v in kwargs.items():
            try:
                # Create figure and axis
                fig, ax = new_fig()

                # Plot transposed array (each row becomes a line)
                ax.plot(v.T)

                # Label axes
                ax.set_xlabel('Index')
                ax.set_ylabel(k)

                # Apply tight layout
                fig.tight_layout()

                # Generate filename and save figure
                filename = f"{k}{self.file_id}.png"
                filename = os.path.join(self.save_dir, filename)
                fig.savefig(filename)

            except Exception:
                # Silently skip any error (consider logging)
                pass


class PlotLines2(Module):
    """
    Plot lines
    """
    create_folder = True

    @debug_decorator
    def __call__(self,
                 default=None,
                 # parameters to skip:
                 terrain_temp=[], terrain_prim=[],
                 last=None, call_number=None,
                 call_total=None, size=None, resolution=None,
                 extent=None, loop_id=None, loop_id_r=None,
                 # plot-parameters
                 exportmode=False, dpi=200, overwrite=False,
                 color=None, cmap=None,
                 grid=False,
                 yscale='linear',
                 **kwargs):

        # Loop over all key-value pairs in kwargs
        for k1, v1 in kwargs.items():
            for k2, v2 in kwargs.items():
                # Skip plotting identical variables or mismatched lengths
                # if k1 == k2 or len(v1) != len(v2):
                if k1 == k2:
                    continue

                # Generate filename and save figure
                filename = f"{k1}_v_{k2}{self.file_id}.png"
                filename = os.path.join(self.save_dir, filename)
                if os.path.isfile(filename) and not overwrite:
                    continue

                try:
                    # Create figure and axis
                    fig, ax = new_fig()

                    if color is not None and color in kwargs:
                        color_data = kwargs[color]
                        print(f"color_data:{color_data}")
                        unique_values = np.unique(color_data)
                        print(f"unique_values:{unique_values}")
                        for v in unique_values:
                            pick = (color_data == v)
                            print(f"pick:{pick}")
                            label = f'{color} = {v:.2f}'
                            print(f"label:{label}")
                            ax.plot(v1[pick], v2[pick], alpha=1, label=label)
                        fig.legend()
                        ax.set_yscale(yscale)
                    else:
                        ax.plot(v1, v2)

                    # Label axes
                    ax.set_xlabel(k1)
                    ax.set_ylabel(k2)

                    # Apply tight layout
                    fig.tight_layout()

                    if grid:
                        ax.grid()
                    fig.savefig(filename)
                    if exportmode:
                        save_all_axes(fig, filename, delta=0.0, dpi=dpi)

                except IndexError as e:
                    # Skip silently on error (could add logging if needed)
                    print(f"e:{e}")
                    pass
                except ValueError as e:
                    # Skip silently on error (could add logging if needed)
                    print(f"e:{e}")
                    pass



import os
import numpy as np
from utils.terrains import Terrain
from utils.utils import Distribution
from utils.plots import save_all_axes
import colorcet as cc
from modules.module import Module, debug_decorator


class Print(Module):
    ''' Print array info '''
    create_folder = False

    @debug_decorator
    def __call__(self, **kwargs):
        for key, value in kwargs.items():
            # if isinstance(value, np.ndarray):
            #     print(f"len(value):{len(value)}")

            # Print
            self.logger.info(f"{key}: {value}")

        # Print some specific information
        self.print_specific(**kwargs)

    def print_specific(self, terrain_prim=None, terrain_temp=None, **_):
        if terrain_prim:
            print(f"len(terrain_prim):{len(terrain_prim)}")
        if terrain_temp:
            print(f"len(terrain_temp):{len(terrain_temp)}")


class Exit(Module):
    ''' '''
    create_folder = False

    @debug_decorator
    def __call__(self, **_):
        exit(0)


class GridSize(Module):
    ''' Test '''
    create_folder = False

    @debug_decorator
    def __call__(self, grid_size=100, default=None, **kwargs):
        grid_size = default if default is not None else grid_size

        if isinstance(grid_size, list):
            grid_size_x, grid_size_y = grid_size
        elif isinstance(grid_size, str) and 'x' in default:
            grid_size_x, grid_size_y = map(float, default.lower().split('x'))
        else:
            grid_size_x, grid_size_y = (grid_size, grid_size)

        grid_size = [grid_size_x, grid_size_y]

        self.info(f"Set grid_size:{grid_size}")

        return {
            'grid_size': grid_size,
            'resolution': None,  # Not valid?
            }


class Size(Module):
    ''' Test '''
    create_folder = False

    @debug_decorator
    def __call__(self, size=50, default=None, **kwargs):
        # göra så att 'default' parametern alltid står först
        # och göra det här automatiskt
        size = default if default is not None else size

        if isinstance(size, list):
            size_x, size_y = size
        elif isinstance(size, str) and 'x' in default:
            size_x, size_y = map(float, default.lower().split('x'))
        else:
            size_x, size_y = (size, size)

        size = [size_x, size_y]

        self.info(f"Set size:{size} [m]")

        return {
            'size': size,
            'extent': [-size[0]/2, size[0]/2, -size[1]/2, size[1]/2],  # test
            }


class Extent(Module):
    ''' Extent '''
    create_folder = False

    @debug_decorator
    def __call__(self, extent=[-25, 25, -25, 25], default=None, **kwargs):
        extent = default if default is not None else extent

        self.info(f"Set extent:{extent} [m]")
        from utils.terrains import extent_to_size

        return {
            'size': extent_to_size(extent),
            'extent': extent,
            }


class Location(Module):
    ''' Update location in extent '''
    create_folder = False

    @debug_decorator
    def __call__(self, location=[0, 0], extent=None, default=None, **kwargs):
        location = default if default is not None else location

        from utils.terrains import extent_to_location
        current_location = extent_to_location(extent)

        # Calculate diff
        diff = [location[0] - current_location[0],
                location[1] - current_location[1]]

        # And shift the extent
        extent = [extent[0]+diff[0],
                  extent[1]+diff[0],
                  extent[2]+diff[1],
                  extent[3]+diff[1]
                  ]

        return {
            'extent': extent,
            }


class Resolution(Module):
    '''  '''
    create_folder = False

    @debug_decorator
    def __call__(self, resolution=5, default=None, **kwargs):
        resolution = default if default is not None else resolution

        self.info(f"Set points-per-meter:{resolution}")

        return {
            'resolution': resolution,
            'grid_size': None,
            }


class Seed(Module):
    """
    Sets a seed

    If `default` is an integer, it overrides `seed`.
    If `default` is 'persistent_random', a random seed is generated once and reused.

    Returns:
        dict: {'seed': resolved seed value}
    """
    create_folder = False

    @debug_decorator
    def __call__(self, seed=0, default=None, **kwargs):
        # Override with default if specified
        if default is not None and isinstance(default, int):
            seed = default
        elif default == 'persistent_random':
            try:
                seed = self.seed
            except AttributeError:
                import random
                seed = random.randint(0, 2**32 - 1)
                self.seed = seed

        self.info(f"Set seed: {seed}")

        return {'seed': seed}


class Folder(Module):
    ''' Set 'folder', affecting e.g Save, and Plot '''
    create_folder = False

    @debug_decorator
    def __call__(self, folder=None, default=None, **kwargs):
        # göra så att 'default' parametern alltid står först
        # och göra det här automatiskt
        folder = default if default is not None else folder

        self.info(f"Set 'folder':{folder}")

        return {
            'folder': folder,
            }


class DebugPlot(Module):
    ''' Do debug plot '''
    create_folder = False

    @debug_decorator
    def __call__(self, default=None, filename='debug_plot.png', **kwargs):
        from utils.plots import debug_plot_horizontal

        filename = default if default is not None else filename
        filename = os.path.join(self.save_dir_original, filename)
        debug_plot_horizontal(filename, **kwargs)


class PlotObstacles(Module):
    ''' Plot obstacles '''
    @debug_decorator
    def __call__(self, size=None, position=[[0, 0]], height=[1], yaw_deg=[0],
                 width=[5], aspect=[1], pitch_deg=[0],
                 filename=None, default=None,
                 exportmode=False, dpi=200,
                 **kwargs):
        # Setup filename
        filename = default if default is not None else filename
        if filename is None:
            filename = f'obstacles{self.file_id}.png'

        filename = os.path.join(self.save_dir, filename)

        from utils.obstacles import Obstacles
        from utils.plots import plot_obstacles
        obstacles = Obstacles()

        obstacles.position = np.array(position).T
        obstacles.height = np.array(height).reshape(1, -1)
        obstacles.yaw_deg = np.array(yaw_deg).reshape(1, -1)
        obstacles.width = np.array(width).reshape(1, -1)
        obstacles.aspect = np.array(aspect).reshape(1, -1)
        obstacles.pitch_deg = np.array(pitch_deg).reshape(1, -1)

        extent = [-size/2, size/2, -size/2, size/2]  # TODO: Temporary fix

        fig, ax = plot_obstacles(
            obstacles, xlim=extent[:2], ylim=extent[:-2],
            color=obstacles.height, cmap=cc.cm.coolwarm,
        )
        fig.savefig(filename)
        if exportmode:
            save_all_axes(fig, filename, delta=0.0, dpi=dpi)
            filename = filename.replace('.png', f'_{0:01d}.png')
            # Possibly use as image texture
            return {'texture': filename}


class SaveObstacles(Module):
    ''' Save obstacles '''
    @debug_decorator
    def __call__(self, size=None, position=[[0, 0]], height=[1], yaw_deg=[0],
                 width=[5], aspect=[1], pitch_deg=[0],
                 filename='obstacles.npz', default=None,
                 overwrite=False,
                 **kwargs):
        # Setup filename
        filename = default if default is not None else filename
        filename = os.path.join(self.save_dir, filename)

        # Skip if already exists, unless 'overwrite'
        if os.path.exists(filename) and not overwrite:
            return

        from utils.obstacles import Obstacles
        obstacles = Obstacles()

        obstacles.position = np.array(position).T
        obstacles.height = np.array(height).reshape(1, -1)
        obstacles.yaw_deg = np.array(yaw_deg).reshape(1, -1)
        obstacles.width = np.array(width).reshape(1, -1)
        obstacles.aspect = np.array(aspect).reshape(1, -1)
        obstacles.pitch_deg = np.array(pitch_deg).reshape(1, -1)

        if '.yaml' in filename:
            obstacles.save_yaml(filename)
        else:
            obstacles.save_numpy(filename)


class Set(Module):
    create_folder = False
    '''
    Set a parameter as Set:parameter=value
    '''
    @debug_decorator
    def __call__(self, default, **_):
        import ast

        # Split the string into key and value
        key, value = default.split('=')
        # Use ast.literal_eval to safely evaluate the value
        try:
            value = ast.literal_eval(value)
        except ValueError:
            pass

        # Create the dictionary
        result_dict = {key: value}

        return result_dict


class SetDistribution(Module):
    create_folder = False
    '''
    Set a distribution as SetDistribution:parameter=distribution(*args)
    '''
    @debug_decorator
    def __call__(self, default, **_):
        # Parse distribution, and return
        from utils.utils import parse_and_assign_distribution
        varname, dist_obj = parse_and_assign_distribution(default)

        return {f'{varname}_distribution': dist_obj}


class Sample(Module):
    create_folder = False
    '''
    Sample a parameter from a distribution
    '''
    @debug_decorator
    def __call__(self, default, **kwargs):
        import ast
        parameter_name = default

        # Get corresponding distribution
        distribution = kwargs[f'{parameter_name}_distribution']

        return {parameter_name: distribution(size=1)}


class SetLogger(Module):
    """Activate and configure the shared logger."""
    create_folder = False

    def __call__(self, default=None, level="info", **_):
        import logging
        # import sys
        logger = logging.getLogger("artificial_terrains")

        level = default if default is not None else level

        # Remove any old handlers to avoid duplicates
        # logger.handlers.clear()

        # Normalize level string → numeric constant
        levels = {
            "debug": logging.DEBUG,
            "info": logging.INFO,
            "warning": logging.WARNING,
            "error": logging.ERROR,
            "critical": logging.CRITICAL,
            "off": logging.CRITICAL + 1,
        }
        level_value = levels.get(str(level).lower(), logging.INFO)
        logger.setLevel(level_value)

        # Create console handler
        # handler = logging.StreamHandler(sys.stdout)
        # formatter = logging.Formatter("%(levelname)s: %(message)s")
        # handler.setFormatter(formatter)
        # logger.addHandler(handler)
        # logger.info(f"Logger activated (level={level.upper()})")


class Random(Module):
    create_folder = False
    '''
    Set random position, height, width, yaw, etc.
    '''
    @debug_decorator
    def __call__(
            self, number_of_values=3,
            size=None, grid_size=None, resolution=None, extent=None,
            position_x_distribution=None,
            position_y_distribution=None,
            height_distribution=Distribution('uniform', 1, 5),
            width_distribution=Distribution('uniform', 2, 10),
            aspect_distribution=Distribution('uniform', 0.5, 1.5),
            yaw_deg_distribution=Distribution('uniform', 0, 360),
            pitch_deg_distribution=Distribution('uniform', 0, 30),
            position_probability_2d=None,
            terrain_temp=[],
            seed=None,
            default=None, **kwargs):

        to_generate = ['position', 'height', 'yaw_deg', 'width', 'aspect',
                       'pitch_deg']

        if seed is not None:
            np.random.seed(seed=seed)

        pipe = {}

        # Parse default input
        if default is not None and isinstance(default, (int, float)):
            number_of_values = default
        if default is not None and isinstance(default, (str)):
            if '[' not in default:
                to_generate = [default]
            else:
                to_generate = default

        print(f"number_of_values:{number_of_values}")
        from utils.artificial_shapes import determine_extent_and_resolution
        extent, (N_x, N_y) = determine_extent_and_resolution(
            resolution, size, grid_size, extent)

        # Setup distributions that depend on other parameters
        if position_x_distribution is None:
            position_x_distribution = Distribution('uniform', extent[0], extent[1])
        if position_y_distribution is None:
            position_y_distribution = Distribution('uniform', extent[2], extent[3])

        # test. Return weights given Random:weights
        if default == 'weights':
            num_octaves = len(terrain_temp)
            weights = self.get_weights(num_octaves)
            return {'weights': weights}

        # Generate position
        if 'position' in to_generate:
            pos_x = position_x_distribution(size=number_of_values)
            pos_y = position_y_distribution(size=number_of_values)
            pipe['position'] = np.stack([pos_x, pos_y], axis=-1)

            # Generate from position_probability_2d if given
            if position_probability_2d is not None:
                pipe['position'] = self.points_from_probability(
                    position_probability_2d, number_of_values)

        name_to_distribution_mapping = {
            'width': width_distribution,
            'height': height_distribution,
            'aspect': aspect_distribution,
            'yaw_deg': yaw_deg_distribution,
            'pitch_deg': pitch_deg_distribution,
        }

        # Loop through parameters, and run their distribution function
        for name, distribution in name_to_distribution_mapping.items():
            if name in to_generate:
                lookup_name = f'{name}_lookup_function'
                if lookup_name in kwargs:
                    # Get values using lookup array and list of positions
                    from utils.interpolator import Interpolator
                    terrain = kwargs[lookup_name]
                    interpolator = Interpolator(terrain.array, terrain.extent)
                    position = pipe['position']
                    sampled_values = interpolator.interpolator(position)
                    pipe[name] = sampled_values

                else:
                    # Simply call the distribution
                    pipe[name] = distribution(size=number_of_values)

        return pipe

    def points_from_probability(self, prob: Terrain, number_of_values):
        from utils.noise import generate_points_from_2D_prob
        # Generate integers between (0, 0) and prob.array.shape
        points = generate_points_from_2D_prob(prob.array, number_of_values)
        # Add uniform random noise in [0, 1)
        points = points + np.random.uniform(size=points.shape)

        # Map to extent
        x = np.interp(points[0], [0, prob.array.shape[0]], prob.extent[:2])
        y = np.interp(points[1], [0, prob.array.shape[1]], prob.extent[2:4])

        # Merge into combined array again
        position = np.array([x, y]).T

        return position

    def get_weights(self, num_octaves, start=128, amplitude_start=10,
                    persistance=0.60, random_amp=0.5, random_sign=True, **_):
        ''' Generate a list of amplitudes '''
        # Generate amplitude list
        amplitude_start_log = np.log(amplitude_start)/np.log(persistance)
        amplitude_end_log = amplitude_start_log+num_octaves-1
        amplitude_list = np.logspace(amplitude_start_log, amplitude_end_log,
                                     num_octaves, base=persistance)

        # Possibly multiply amplitudes by random perturbations
        if random_amp:
            random_factor = np.random.normal(
                loc=1.0, scale=random_amp, size=amplitude_list.shape)
            amplitude_list *= random_factor
        if random_sign:
            random_sign = np.random.choice([-1, 1], size=amplitude_list.shape)
            amplitude_list *= random_sign

        return amplitude_list


class LoadObstacles(Module):
    ''' '''
    create_folder = False

    @debug_decorator
    def __call__(self,
                 filename='obstacles.npz',
                 extent=None,
                 default=None,
                 call_number=None,
                 call_total=None,
                 exportmode=None,
                 overwrite=False,
                 **pipe):
        filename = default if default is not None else filename

        from utils.obstacles import Obstacles

        if 'yaml' in filename:
            obstacles = Obstacles.from_yaml(filename)
        else:
            obstacles = Obstacles.from_numpy(filename)

        # Filter
        # pick = (obstacles.height > 0.29).squeeze()
        # obstacles = obstacles[pick]

        # Add to pipe, in same manner as the 'Random' module does

        pipe['position'] = obstacles.position.T
        # For these, we remove one dimension
        pipe['width'] = obstacles.width[0, :]
        pipe['height'] = obstacles.height[0, :]
        pipe['yaw_deg'] = obstacles.yaw_deg[0, :]
        pipe['aspect'] = obstacles.aspect[0, :]
        pipe['pitch_deg'] = obstacles.pitch_deg[0, :]

        return pipe


class RemoveDistantObstacles(Module):
    """
    Filter obstacles based on their distance

    Parameters:
    - distance_limit (float): The maximum distance an obstacle can be from
      the extent to be kept. Obstacles beyond this limit are removed.
    - extent (list): A list or tuple defining the bounds of the region in
      which obstacles should be kept. Should be in the format [xmin, xmax, ymin, ymax].
    - default (optional): If set, it overrides the default distance_limit.
    """
    create_folder = False

    @debug_decorator
    def __call__(self,
                 distance_limit=100,
                 extent=None,
                 default=None,
                 **pipe):
        from utils.terrains import distances_to_extent
        # Use the provided default distance_limit if set, otherwise use the
        # given value for distance_limit.
        distance_limit = default if default is not None else distance_limit

        # Calculate the closest distance to the extent for all positions
        distances = distances_to_extent(pipe['position'], extent)

        # Determine which obstacles are within the distance limit
        keep = (distances <= distance_limit)

        result = {}
        # Filter the obstacle data based on the 'keep' boolean mask
        for name in [
                'position', 'width', 'height',
                'yaw_deg', 'aspect', 'pitch_deg']:
            result[name] = pipe[name][keep]

        return result



from modules.module import Module, debug_decorator
import numpy as np
import os
from utils.utils import get_terrains, get_terrain


class Negate(Module):
    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=[], terrain_prim=[],
                 default=None, last=None, **_):
        terrains = get_terrains(
            terrain_temp, terrain_prim, last, remove=False,
            print_fn=self.info)

        # Negate the terrains
        for terrain in terrains:
            terrain.array = -terrain.array


class Scale(Module):
    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=[], terrain_prim=[],
                 factor=2, default=None, last=None, **_):
        factor = default if default is not None else factor

        terrains = get_terrains(
            terrain_temp, terrain_prim, last, remove=False,
            print_fn=self.info)

        # Scale the terrains
        for terrain in terrains:
            terrain.array *= factor


class Add(Module):
    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=[], terrain_prim=[],
                 term=1, default=None, last=None, **_):
        term = default if default is not None else term

        terrains = get_terrains(
            terrain_temp, terrain_prim, last, remove=False,
            print_fn=self.info)

        # Scale the terrains
        for terrain in terrains:
            terrain.array += term


class Absolute(Module):
    ''' Apply the absolute operation on the terrain '''

    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=[], terrain_prim=[],
                 default=None, last=None, **_):
        terrains = get_terrains(
            terrain_temp, terrain_prim, last, remove=False,
            print_fn=self.info)

        # Scale the terrains
        for terrain in terrains:
            terrain.array = np.absolute(terrain.array)


class Clip(Module):
    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=[], terrain_prim=[],
                 clip_min=0, clip_max=1,
                 default=None, last=None, **_):
        ''' Clip terrains '''

        terrains = get_terrains(
            terrain_temp, terrain_prim, last, remove=False,
            print_fn=self.info)

        # Clip
        for terrain in terrains:
            terrain.array = np.clip(terrain.array, clip_min, clip_max)


class Around(Module):
    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=[], terrain_prim=[],
                 around=1, default=None, last=None, **_):
        ''' Move terrain in z to place mean at some value '''
        # Get value from 'default'
        around = default if default is not None else around

        terrains = get_terrains(
            terrain_temp, terrain_prim, last, remove=False,
            print_fn=self.info)

        # Move to place mean around 1
        for terrain in terrains:
            terrain.array = terrain.array + around - np.nanmean(terrain.array)


class Around1(Around):
    def __call__(self, *args, around=None, **kwargs):
        super().__call__(*args, around=1, **kwargs)


class AsProbability(Module):
    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=[], terrain_prim=[],
                 default=None, last=None, **_):
        ''' Use terrain as a 2d probability '''
        # Get and remove terrain from dict/heap
        terrain = get_terrain(
            terrain_temp, terrain_prim, print_fn=self.info)

        # Clip to remove negative values
        terrain.array = np.clip(terrain.array, 0, None)
        # Normalize
        terrain.array = terrain.array/np.sum(terrain.array)

        return {'position_probability_2d': terrain}


class AsLookupFor(Module):
    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=[], terrain_prim=[],
                 parameter='width',
                 default=None, last=None, **_):
        ''' Use terrain as function f(x, y) to sample values given positions '''
        # Use default input for the parameter
        parameter = default if default is not None else parameter

        # Get and remove terrain from dict/heap
        terrain = get_terrain(
            terrain_temp, terrain_prim, print_fn=self.info)

        return {f'{parameter}_lookup_function': terrain}


class AsFactor(Module):
    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=[], terrain_prim=[],
                 default=None, last=None, **_):
        ''' Use terrain as a 2d factor '''
        # Get and remove terrain from dict/heap
        terrain = get_terrain(
            terrain_temp, terrain_prim, print_fn=self.info, remove=True)

        return {'factor': terrain.array}


class BezierRemap(Module):
    ''' Nonlinear remapping of values using bezier curves

    Uses a 3rd order bezier curve, with p0=(0, 0) and p3=(1, 1).
    p1, and p2 are free, and are used to shape the bezier curve.
    p1 and p2 are set given 1, 2, or 4 input numbers.

    Args:
      bezier_args: (float, or list of length 1, 2, or 4)
        a float in [-1, 1]:  p1=p2=[a, 1-a]
        [a, b] floats in [-1, 1]:  p1=p2=[a, b]
        [a, b, c, d] floats in [-1, 1]:  p1=[a, b], p2=[c, d]

    '''
    create_folder = True

    @debug_decorator
    def __call__(self, bezier_args=0.5, terrain_temp=[], terrain_prim=[],
                 default=None, plot=False, call_number=None, call_total=None,
                 last=None, **pipe):

        bezier_args = default if default is not None else bezier_args
        # Make sure a is a list
        if not isinstance(bezier_args, list):
            bezier_args = [bezier_args]

        terrains = get_terrains(
            terrain_temp, terrain_prim, last, remove=False,
            print_fn=self.info)

        # Get bezier curve
        t_array = np.linspace(0, 1, 50)
        xp, fp, points = self.bezier_function(t_array, *bezier_args)

        # plot
        if plot:
            from utils.plots import new_fig
            fig, ax = new_fig()
            ax.plot(xp, fp)
            filename = os.path.join(self.save_dir, 'bezier.png')
            ax.scatter(*points.T)
            for i, point in enumerate(points):
                ax.annotate(i, point)
            fig.savefig(filename)

        # Apply nonlinear bezier remapping
        for terrain in terrains:
            # Normalize array
            z_min = np.min(terrain.array)
            z_max = np.max(terrain.array)
            normalized_array = (terrain.array - z_min) / (z_max - z_min)

            # Use bezier curve to do nonlinear remapping
            transformed_array = np.interp(normalized_array, xp, fp)

            # Rescale to original limits
            rescaled_array = transformed_array * (z_max - z_min) + z_min

            # Assign to terrain
            terrain.array = rescaled_array

    @staticmethod
    def bezier_function(t, *args, p0=[0, 0], p1=None, p2=None, p3=[1, 1]):
        if p1 is None or p2 is None:
            # Set from a
            if len(args) == 1:
                a = args[0]
                p1 = [a, 1-a]
                p2 = [a, 1-a]

            elif len(args) == 2:
                a, b = args
                p1 = [a, b]
                p2 = [a, b]
            elif len(args) == 4:
                a, b, c, d = args
                p1 = [a, b]
                p2 = [c, d]

        p0 = np.array(p0).reshape(-1, 2)
        p1 = np.array(p1).reshape(-1, 2)
        p2 = np.array(p2).reshape(-1, 2)
        p3 = np.array(p3).reshape(-1, 2)

        t = t.reshape(-1, 1)

        curve = (1-t)**3*p0 + 3*(1-t)**2*t*p1 + 3*(1-t)*t**2*p2 + t**3*p3
        xp = curve[:, 0]
        fp = curve[:, 1]

        points = [p0.squeeze(), p1.squeeze(), p2.squeeze(), p3.squeeze()]

        return np.array(xp), np.array(fp), np.array(points)


class Smooth(Module):
    """
    Perform Gaussian smoothing
    """
    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=[], terrain_prim=[],
                 default=None, last=None, sigma_meter=5, **_):
        from scipy.ndimage import gaussian_filter

        # Possibly set sigma (m) using default value
        sigma_meter = default if default is not None else sigma_meter
        self.info(f"smooth using sigma: {sigma_meter} (m)")

        # Get terrains
        terrains = get_terrains(
            terrain_temp, terrain_prim, last, remove=False,
            print_fn=self.info)

        # Loop the terrains, and perform Gaussian smoothing
        for terrain in terrains:
            resolution = terrain.resolution  # (dy, dx)
            sigma_grid = np.divide(sigma_meter, resolution)

            terrain.array = gaussian_filter(terrain.array, sigma=sigma_grid)


class Resample(Module):
    """
    Resample the resolution (array shape) of terrains, keeping the same extent.

    Parameters
    ----------
    new_shape : tuple[int, int] | None
        (new_nx, new_ny). If None, `scale` is used instead.
    scale : float | tuple[float, float] | None
        Scale factor(s) applied to (nx, ny). For example:
        - 2.0  -> double resolution in both directions
        - (2.0, 1.0) -> double in x, keep y
    """
    create_folder = False

    @debug_decorator
    def __call__(self, terrain_temp=[], terrain_prim=[],
                 new_shape=None,
                 resolution=None,
                 default=None,
                 last=None,
                 **_):
        # Let "default" act as fallback scale if user wires it that way
        if default is not None and isinstance(default, float):
            # interpret as "resolution" (points per meter)
            resolution = default
        elif default is not None and isinstance(default, int):
            # interpret as new array shape
            new_shape = default

        # Get last terrain terrain
        terrain = get_terrain(
            terrain_temp, terrain_prim, print_fn=self.info,
            remove=False)

        hf_array = terrain.array
        extent = terrain.extent  # (left, right, bottom, top)
        size = terrain.size
        nx, ny = hf_array.shape

        if new_shape is not None:
            new_nx, new_ny = new_shape
        else:
            # TODO: Take into account that resolution could be different for x and y
            new_nx = int(resolution * size[0])
            new_ny = int(resolution * size[1])

        # --- Build interpolator from current terrain ------------------
        from utils.interpolator import Interpolator
        interp = Interpolator(hf_array, extent)  # uses RegularGridInterpolator internally

        left, right, bottom, top = extent

        # New grid in same world extent, new resolution
        new_x = np.linspace(left,  right,  new_nx)
        new_y = np.linspace(bottom, top,   new_ny)

        # Be consistent with Interpolator: first dim = x, second dim = y
        X, Y = np.meshgrid(new_x, new_y, indexing='ij')  # (new_nx, new_ny)

        pts = np.stack([X.ravel(), Y.ravel()], axis=-1)  # (N, 2) with cols (x, y)

        new_heights = interp.interpolator(pts).reshape((new_nx, new_ny))

        # --- Write back to terrain -----------------------------------
        terrain.array = new_heights
